# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Custom LiteLLM logger that writes token/cost usage and request/response trace to a JSONL file.

Environment variables:
    EXGENTIC_LLM_LOG_FILE: optional override of the output JSONL path (default: trace.jsonl in CWD).
    EXGENTIC_OTEL_ENABLED: enable OpenTelemetry span creation for LLM calls.
"""

from __future__ import annotations

import json
import os
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from litellm.integrations.custom_logger import CustomLogger

from ...utils.settings import get_settings

# Environment variable constants
FILE_ENV = "EXGENTIC_LLM_LOG_FILE"
DEFAULT_FILE = "trace.jsonl"


def _otel_enabled() -> bool:
    """Check if OTEL is enabled via settings."""
    return bool(get_settings().otel_enabled)


def _otel_record_content() -> bool:
    """Check if OTEL content recording is enabled via settings."""
    return get_settings().otel_record_content


class TraceLogger(CustomLogger):
    def __init__(self, file_path: str | None = None) -> None:
        super().__init__()
        self._file_path = file_path
        self._tracer = None
        self._otel_logger = None
        self._context = None

    @staticmethod
    def _ensure_context() -> None:
        try:
            from ...core.context import init_context_from_env, try_get_context

            if try_get_context() is not None:
                return

            init_context_from_env()
        except RuntimeError:
            pass

    def _init_otel(self, kwargs) -> None:
        import threading

        from ...utils.otel import get_session_logger, init_tracing_from_env

        ctx = self.get_context(kwargs)
        if ctx is None or ctx.session_id is None or ctx.otel_context is None:
            # no otel context means tracing cannot be initialized
            warnings.warn(
                f"No OTEL context found for TraceLogger, skipping OTEL initialization. context={ctx}",
                stacklevel=2,
            )
            return

        # Initialize the TracerProvider (use simple processor for subprocess)
        self._tracer = init_tracing_from_env()

        base = Path(ctx.output_dir) / ctx.run_id
        session_root = base / "sessions" / ctx.session_id
        self._otel_logger = get_session_logger(
            session_root,
            f"{__name__} | pid={os.getpid()} tid={threading.get_native_id()}",
        )

    def _get_parent_context(self, kwargs) -> Any:
        """Reconstruct parent span context from environment variables or ContextVar.

        For proxy subprocess: reads from environment variables
        For direct callback: reads from Context ContextVar
        """
        from opentelemetry import context, trace
        from opentelemetry.trace import NonRecordingSpan, SpanContext, TraceFlags

        ctx = self.get_context(kwargs)

        trace_id_hex = ctx.otel_context.trace_id
        span_id_hex = ctx.otel_context.span_id

        if not trace_id_hex or not span_id_hex:
            return context.get_current()

        trace_id = int(trace_id_hex, 16)
        span_id = int(span_id_hex, 16)

        self._otel_logger.log_context_read(trace_id_hex, span_id_hex)

        span_context = SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            is_remote=True,
            trace_flags=TraceFlags(0x01),
        )

        parent_span = NonRecordingSpan(span_context)
        return trace.set_span_in_context(parent_span)

    def _create_llm_span(self, kwargs, name: str, start_time: Optional[Any] = None) -> Optional[Any]:
        """Create span for LLM call with CLIENT span kind.

        Args:
            kwargs: Keyword arguments containing context and other parameters
            name: Name of the span
            start_time: Optional datetime when the span started
        """
        from opentelemetry.trace import SpanKind

        parent_ctx = self._get_parent_context(kwargs)

        if start_time:
            # Convert datetime to nanoseconds since epoch for OTEL
            start_time_ns = int(start_time.timestamp() * 1_000_000_000)
            span = self._tracer.start_span(name, context=parent_ctx, start_time=start_time_ns, kind=SpanKind.CLIENT)
        else:
            span = self._tracer.start_span(name, context=parent_ctx, kind=SpanKind.CLIENT)

        span_ctx = span.get_span_context()

        # Extract parent span ID from context for logging
        ctx = self.get_context(kwargs)
        parent_span_id = ctx.otel_context.span_id if ctx and ctx.otel_context else None

        self._otel_logger.log_span_start(
            span_name=name,
            span_id=format(span_ctx.span_id, "016x"),
            trace_id=format(span_ctx.trace_id, "032x"),
            parent_span_id=parent_span_id,
            start_time=start_time,
        )

        return span

    def _set_attribute(self, span: Any, key: str, value: Any) -> None:
        span.set_attribute(key, value)
        span_ctx = span.get_span_context()
        self._otel_logger.log_attribute_set(key, value, format(span_ctx.span_id, "016x"))

    @staticmethod
    def _metadata_context(kwargs: dict[str, Any]):
        from ...core.context import Context, OtelContext, Role

        # Check multiple locations where metadata might be stored
        # 1. Direct litellm_metadata parameter
        metadata = kwargs.get("litellm_metadata")
        if not metadata:
            # 2. litellm_params.litellm_metadata
            metadata = kwargs.get("litellm_params", {}).get("litellm_metadata")
        if not metadata:
            # 3. litellm_params.metadata (for 'metadata' parameter)
            metadata = kwargs.get("litellm_params", {}).get("metadata")

        if not isinstance(metadata, dict):
            return None

        context = metadata.get("context")
        if isinstance(context, Context):
            return context

        # Reconstruct Context from serialized fields
        # Check if we have the required fields
        if "exgentic_ctx_run_id" not in metadata:
            return None

        # Reconstruct OtelContext if present
        otel_context = None
        if "exgentic_ctx_otel_trace_id" in metadata and "exgentic_ctx_otel_span_id" in metadata:
            otel_context = OtelContext(
                trace_id=metadata["exgentic_ctx_otel_trace_id"],
                span_id=metadata["exgentic_ctx_otel_span_id"],
            )

        # Reconstruct Role
        role_str = metadata.get("exgentic_ctx_role", "framework")
        try:
            role = Role(role_str)
        except ValueError:
            role = Role.FRAMEWORK

        # Reconstruct Context
        return Context(
            run_id=metadata["exgentic_ctx_run_id"],
            output_dir=metadata["exgentic_ctx_output_dir"],
            cache_dir=metadata["exgentic_ctx_cache_dir"],
            session_id=metadata.get("exgentic_ctx_session_id"),
            task_id=metadata.get("exgentic_ctx_task_id"),
            role=role,
            otel_context=otel_context,
        )

    @staticmethod
    def _context_log_path(ctx) -> str:
        base = Path(ctx.output_dir) / ctx.run_id
        if ctx.session_id:
            return str(base / "sessions" / ctx.session_id / ctx.role.value / "litellm" / "trace.jsonl")
        return str(base / "run" / "litellm" / "trace.jsonl")

    def get_context(self, kwargs: dict[str, Any]):
        from ...core.context import Context, try_get_context

        self._ensure_context()
        metadata_context = self._metadata_context(kwargs)
        if isinstance(metadata_context, Context):
            return metadata_context
        context = kwargs.get("context")
        if isinstance(context, Context):
            return context
        return try_get_context()

    def _resolve_log_path(self, kwargs: dict[str, Any]) -> str:
        from ...core.context import Context

        if self._file_path:
            return self._file_path

        self._ensure_context()
        metadata_context = self._metadata_context(kwargs)
        if metadata_context is not None:
            return self._context_log_path(metadata_context)

        context = kwargs.get("context")
        if isinstance(context, Context):
            return self._context_log_path(context)

        file_path = os.environ.get(FILE_ENV)
        if file_path:
            return file_path

        ctx = self.get_context(kwargs)  # TODO: this is redundant
        if ctx is not None:
            return self._context_log_path(ctx)

        return DEFAULT_FILE

    def _write_row(self, kwargs: dict[str, Any], response_obj: dict[str, Any], status: str) -> None:
        file_path = self._resolve_log_path(kwargs)
        Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        usage = self._extract_usage(response_obj)
        row = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "status": status,
            "model": kwargs.get("model"),
            "prompt_tokens": usage.get("prompt_tokens"),
            "completion_tokens": usage.get("completion_tokens"),
            "total_tokens": usage.get("total_tokens"),
            "cost": kwargs.get("response_cost"),
            "request": self._capture_request(kwargs),
            "response": self._capture_response(response_obj),
            "trace_id": kwargs.get("litellm_trace_id") or kwargs.get("litellm_call_id"),
        }

        with open(file_path, "a", encoding="utf-8") as f:
            f.write(json.dumps(row, ensure_ascii=False, default=str) + "\n")

    def _write_otel(
        self,
        kwargs: dict[str, Any],
        response_obj: dict[str, Any],
        status: str,
        start_time=None,
        end_time=None,
    ) -> None:
        """Write OTEL span for LLM call following GenAI inference span conventions.

        Follows OpenTelemetry GenAI semantic conventions for inference spans:
        https://opentelemetry.io/docs/specs/semconv/gen-ai/gen-ai-spans/#inference

        Span name format: {gen_ai.operation.name} {gen_ai.request.model}
        Span kind: CLIENT
        """
        try:
            if not _otel_enabled():
                return

            # Lazy initialize OTEL if not already done
            if self._tracer is None:
                self._init_otel(kwargs)
            ctx = self.get_context(kwargs)

            # Lazy import GenAI semantic conventions
            from opentelemetry.trace import Status, StatusCode

            optional_params = kwargs.get("optional_params", {}) or {}
            litellm_params = kwargs.get("litellm_params", {}) or {}

            # ===== DETERMINE OPERATION TYPE =====
            # Required attribute: gen_ai.operation.name
            operation = "chat" if kwargs.get("messages") else "text_completion"

            # ===== CREATE SPAN WITH PROPER NAME =====
            # Span name format: {gen_ai.operation.name} {gen_ai.request.model}
            model = kwargs.get("model", "unknown")
            span_name = f"{operation} {model}"
            span = self._create_llm_span(kwargs, span_name, start_time=start_time)

            # ===== SET SPAN STATUS =====
            if status == "success":
                span.set_status(Status(StatusCode.OK))
            else:
                span.set_status(Status(StatusCode.ERROR))
                # Conditionally Required: error.type if operation ended in error
                error_info = kwargs.get("exception") or response_obj.get("error")
                if error_info:
                    if isinstance(error_info, dict):
                        error_type = error_info.get("type") or error_info.get("code") or "unknown_error"
                    elif isinstance(error_info, Exception):
                        error_type = type(error_info).__name__
                    else:
                        error_type = str(error_info)
                    self._set_attribute(span, "error.type", error_type)

            # ===== REQUIRED ATTRIBUTES =====
            # gen_ai.operation.name (Required)
            self._set_attribute(span, "gen_ai.operation.name", operation)

            # gen_ai.provider.name (Required) - maps LiteLLM provider to standard names
            provider = litellm_params.get("custom_llm_provider", "unknown")
            if provider is None:
                provider = "unknown"
            # Map common LiteLLM providers to standard GenAI provider names
            provider_mapping = {
                "openai": "openai",
                "azure": "azure.ai.openai",
                "anthropic": "anthropic",
                "bedrock": "aws.bedrock",
                "vertex_ai": "gcp.vertex_ai",
                "gemini": "gcp.gemini",
                "cohere": "cohere",
                "groq": "groq",
                "mistral": "mistral_ai",
                "deepseek": "deepseek",
                "perplexity": "perplexity",
                "watsonx": "ibm.watsonx.ai",
                "xai": "x_ai",
            }
            standard_provider = provider_mapping.get(provider.lower(), provider)
            self._set_attribute(span, "gen_ai.provider.name", standard_provider)

            # ===== CONDITIONALLY REQUIRED ATTRIBUTES =====
            # gen_ai.request.model (Conditionally Required if available)
            if model:
                self._set_attribute(span, "gen_ai.request.model", model)

            # gen_ai.request.choice.count (Conditionally Required if available and !=1)
            n = optional_params.get("n")
            if n is not None and n != 1:
                self._set_attribute(span, "gen_ai.request.choice.count", n)

            # gen_ai.request.seed (Conditionally Required if applicable and request includes seed)
            seed = optional_params.get("seed")
            if seed is not None:
                self._set_attribute(span, "gen_ai.request.seed", seed)

            # server.port (Conditionally Required if server.address is set)
            # Note: LiteLLM doesn't typically expose server details, skip for now

            # ===== RECOMMENDED REQUEST ATTRIBUTES =====
            self._set_attribute(span, "gen_ai.conversation.id", ctx.session_id)

            if optional_params.get("max_tokens") is not None:
                self._set_attribute(span, "gen_ai.request.max_tokens", optional_params["max_tokens"])

            if optional_params.get("temperature") is not None:
                self._set_attribute(span, "gen_ai.request.temperature", optional_params["temperature"])

            if optional_params.get("top_p") is not None:
                self._set_attribute(span, "gen_ai.request.top_p", optional_params["top_p"])

            if optional_params.get("top_k") is not None:
                self._set_attribute(span, "gen_ai.request.top_k", float(optional_params["top_k"]))

            if optional_params.get("frequency_penalty") is not None:
                self._set_attribute(
                    span,
                    "gen_ai.request.frequency_penalty",
                    optional_params["frequency_penalty"],
                )

            if optional_params.get("presence_penalty") is not None:
                self._set_attribute(
                    span,
                    "gen_ai.request.presence_penalty",
                    optional_params["presence_penalty"],
                )

            # gen_ai.request.stop_sequences (Recommended)
            stop = optional_params.get("stop")
            if stop is not None:
                if isinstance(stop, list):
                    self._set_attribute(span, "gen_ai.request.stop_sequences", stop)
                else:
                    self._set_attribute(span, "gen_ai.request.stop_sequences", [stop])

            # ===== RECOMMENDED RESPONSE ATTRIBUTES =====
            if response_obj:
                # Helper function to safely get attribute from dict or object
                def safe_get(obj, key, default=None):
                    if isinstance(obj, dict):
                        return obj.get(key, default)
                    return getattr(obj, key, default)

                # Get span context for logging
                span_ctx = span.get_span_context()
                span_id_hex = format(span_ctx.span_id, "016x")

                # gen_ai.response.id (Recommended)
                response_id = safe_get(response_obj, "id")
                if response_id:
                    self._set_attribute(span, "gen_ai.response.id", response_id)
                    if self._otel_logger:
                        self._otel_logger.log_attribute_set("gen_ai.response.id", response_id, span_id_hex)

                # gen_ai.response.model (Recommended)
                response_model = safe_get(response_obj, "model")
                if response_model:
                    self._set_attribute(span, "gen_ai.response.model", response_model)
                    if self._otel_logger:
                        self._otel_logger.log_attribute_set("gen_ai.response.model", response_model, span_id_hex)

                # gen_ai.usage.input_tokens and gen_ai.usage.output_tokens (Recommended)
                usage = safe_get(response_obj, "usage")
                if usage:
                    prompt_tokens = safe_get(usage, "prompt_tokens")
                    if prompt_tokens is not None:
                        self._set_attribute(span, "gen_ai.usage.input_tokens", prompt_tokens)
                        if self._otel_logger:
                            self._otel_logger.log_attribute_set("gen_ai.usage.input_tokens", prompt_tokens, span_id_hex)

                    completion_tokens = safe_get(usage, "completion_tokens")
                    if completion_tokens is not None:
                        self._set_attribute(span, "gen_ai.usage.output_tokens", completion_tokens)
                        if self._otel_logger:
                            self._otel_logger.log_attribute_set(
                                "gen_ai.usage.output_tokens",
                                completion_tokens,
                                span_id_hex,
                            )

                # gen_ai.response.finish_reasons (Recommended)
                choices = safe_get(response_obj, "choices", [])
                if choices:
                    finish_reasons = []
                    for choice in choices:
                        finish_reason = safe_get(choice, "finish_reason")
                        if finish_reason:
                            finish_reasons.append(finish_reason)
                    if finish_reasons:
                        self._set_attribute(span, "gen_ai.response.finish_reasons", finish_reasons)
                        if self._otel_logger:
                            self._otel_logger.log_attribute_set(
                                "gen_ai.response.finish_reasons",
                                finish_reasons,
                                span_id_hex,
                            )

            # ===== OPT-IN ATTRIBUTES (for content recording) =====
            # Note: These are opt-in and may contain sensitive data
            # Only include if explicitly enabled via settings
            if _otel_record_content():
                # gen_ai.tool.definitions (Opt-In)
                tools = kwargs.get("tools")
                if tools:
                    try:
                        self._set_attribute(
                            span,
                            "gen_ai.tool.definitions",
                            json.dumps(tools, default=str),
                        )
                    except Exception:
                        pass  # Skip if serialization fails

                # gen_ai.input.messages (Opt-In) - structured format
                messages = kwargs.get("messages")
                if messages:
                    # Convert to GenAI message format
                    try:
                        self._set_attribute(
                            span,
                            "gen_ai.input.messages",
                            json.dumps(messages, default=str),
                        )
                    except Exception:
                        pass  # Skip if serialization fails

                # gen_ai.output.messages (Opt-In) - structured format
                if response_obj:
                    # Helper function already defined above in the response attributes section
                    def safe_get(obj, key, default=None):
                        if isinstance(obj, dict):
                            return obj.get(key, default)
                        return getattr(obj, key, default)

                    choices = safe_get(response_obj, "choices", [])
                    if choices:
                        output_messages = []
                        for choice in choices:
                            message = safe_get(choice, "message")
                            if message:
                                output_msg = {
                                    "role": safe_get(message, "role", "assistant"),
                                    "parts": [],
                                }
                                content = safe_get(message, "content")
                                if content:
                                    output_msg["parts"].append({"type": "text", "content": content})
                                # Include tool calls if present
                                tool_calls = safe_get(message, "tool_calls")
                                if tool_calls:
                                    for tc in tool_calls:
                                        func = safe_get(tc, "function", {})
                                        output_msg["parts"].append(
                                            {
                                                "type": "tool_call",
                                                "id": safe_get(tc, "id"),
                                                "name": safe_get(func, "name"),
                                                "arguments": safe_get(func, "arguments"),
                                            }
                                        )
                                finish_reason = safe_get(choice, "finish_reason")
                                if finish_reason:
                                    output_msg["finish_reason"] = finish_reason
                                output_messages.append(output_msg)

                        if output_messages:
                            try:
                                self._set_attribute(
                                    span,
                                    "gen_ai.output.messages",
                                    json.dumps(output_messages, default=str),
                                )
                            except Exception:
                                pass  # Skip if serialization fails

            # ===== END SPAN =====
            if end_time:
                end_time_ns = int(end_time.timestamp() * 1_000_000_000)
                span.end(end_time=end_time_ns)
            else:
                span.end()

            span_ctx = span.get_span_context()
            if self._otel_logger:
                self._otel_logger.log_span_end(
                    span_name=span_name,
                    span_id=format(span_ctx.span_id, "016x"),
                    status=status,
                    end_time=end_time,
                )
        except Exception:
            pass

    def log_success_event(self, kwargs: dict[str, Any], response_obj: dict[str, Any], start_time, end_time):
        self._write_row(kwargs, response_obj, status="success")
        self._write_otel(
            kwargs,
            response_obj,
            status="success",
            start_time=start_time,
            end_time=end_time,
        )

    async def async_log_success_event(self, kwargs: dict[str, Any], response_obj: dict[str, Any], start_time, end_time):
        self._write_row(kwargs, response_obj, status="success")
        self._write_otel(
            kwargs,
            response_obj,
            status="success",
            start_time=start_time,
            end_time=end_time,
        )

    def log_failure_event(self, kwargs: dict[str, Any], response_obj: dict[str, Any], start_time, end_time):
        self._write_row(kwargs, response_obj, status="failure")
        self._write_otel(
            kwargs,
            response_obj,
            status="failure",
            start_time=start_time,
            end_time=end_time,
        )

    async def async_log_failure_event(self, kwargs: dict[str, Any], response_obj: dict[str, Any], start_time, end_time):
        self._write_row(kwargs, response_obj, status="failure")
        self._write_otel(
            kwargs,
            response_obj,
            status="failure",
            start_time=start_time,
            end_time=end_time,
        )

    async def async_log_stream_event(self, kwargs: dict[str, Any], response_obj: dict[str, Any], start_time, end_time):
        self._write_row(kwargs, response_obj, status="stream")
        self._write_otel(kwargs, response_obj, status="stream")

    # Helpers -----------------------------------------------------

    def _extract_usage(self, response_obj: Any) -> dict[str, Any]:
        usage: dict[str, Any] = {}
        if isinstance(response_obj, dict):
            usage = response_obj.get("usage", {}) or {}
        else:
            usage_attr = getattr(response_obj, "usage", None)
            if isinstance(usage_attr, dict):
                usage = usage_attr
            elif hasattr(usage_attr, "__dict__"):
                usage = usage_attr.__dict__
        return usage

    def _capture_request(self, kwargs: dict[str, Any]) -> dict[str, Any]:
        safe: dict[str, Any] = {}
        for key in (
            "messages",
            "prompt",
            "tools",
            "tool_choice",
            "functions",
            "function_call",
            "temperature",
            "model",
        ):
            if key in kwargs:
                safe[key] = kwargs[key]
        return safe

    def _capture_response(self, response_obj: Any) -> dict[str, Any]:
        if isinstance(response_obj, dict):
            resp = dict(response_obj)
            resp.pop("usage", None)
            return resp
        out: dict[str, Any] = {}
        for attr in ("choices", "id", "object", "created", "model", "error"):
            if hasattr(response_obj, attr):
                out[attr] = getattr(response_obj, attr)
        if not out:
            out["repr"] = repr(response_obj)
        return out


# Sync logger for direct SDK calls (e.g. tool-calling agent).
class SyncTraceLogger(TraceLogger):
    def __call__(self, *args, **kwargs):
        return None


class AsyncTraceLogger(TraceLogger):
    async def __call__(self, *args, **kwargs):
        return None


# Expose module-level instances.
trace_logger = TraceLogger()
sync_trace_logger = SyncTraceLogger()
async_trace_logger = AsyncTraceLogger()



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Shared fixtures for runner/transport tests.

The ``Calculator`` class and the parametrized ``calc`` fixture are used
across all transport test modules so that every transport is verified
against the exact same behavioural contract.
"""

from __future__ import annotations

import pytest
from exgentic.adapters.runners import with_runner
from exgentic.testing.calculator import Calculator, CalculatorError

# Re-export so existing test imports keep working.
__all__ = ["Calculator", "CalculatorError"]

# Runners available for the current milestone.
_AVAILABLE_RUNNERS = ["direct", "thread", "process", "service"]


@pytest.fixture(params=_AVAILABLE_RUNNERS)
def runner_name(request):
    return request.param


@pytest.fixture
def calc(runner_name):
    proxy = with_runner(Calculator, runner=runner_name, value=10)
    yield proxy
    try:
        proxy.close()
    except Exception:
        pass



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
from collections import Counter
from logging import Logger
from typing import Any, Callable, Optional

from pydantic import BaseModel, ValidationError

from ..observers.logging import get_disabled_logger
from .types import (
    Action,
    ActionType,
    MultiObservation,
    SingleAction,
    SingleObservation,
    ValidationReport,
)

ActionHandler = Callable[[SingleAction], Optional[Any]]


def _parse_arguments_payload(arguments: Any) -> Any:
    """Return arguments with best-effort JSON parsing for string inputs."""
    if isinstance(arguments, str):
        try:
            return json.loads(arguments)
        except json.JSONDecodeError:
            return arguments
    return arguments


def _validation_report_error(report: ValidationReport) -> Optional[str]:
    """Return a unified error string if the validation report marks the action invalid."""
    if not report.name_valid:
        return report.error or "Invalid action name"
    if not report.args_valid or not report.valid:
        return report.error or "Invalid arguments"
    return None


def build_action(action_type: ActionType, arguments: Any, *, action_id: Optional[str] = None) -> SingleAction:
    """Best-effort construction of a SingleAction with validity flag and optional ID."""
    parsed_args = _parse_arguments_payload(arguments)

    data: dict[str, Any] = {"name": action_type.name, "arguments": parsed_args}
    if action_id is not None:
        data["id"] = action_id

    report = ValidationReport()
    try:
        action = action_type.cls.model_validate(data)
    except ValidationError as exc:
        report.valid = False
        report.args_valid = False
        report.error = format_validation_errors(exc)
        report.details = {"errors": exc.errors()}
        args_cls = action_type.arguments
        if isinstance(parsed_args, dict) and isinstance(args_cls, type) and issubclass(args_cls, BaseModel):
            try:
                parsed_args = args_cls.model_validate(parsed_args)
                data["arguments"] = parsed_args
            except ValidationError:
                try:
                    parsed_args = args_cls.model_construct(**parsed_args)
                    data["arguments"] = parsed_args
                except Exception:
                    pass
        action = action_type.cls.model_construct(**data)

    # Attach validation metadata on the action instance
    try:
        action.validation = report  # type: ignore[attr-defined]
    except Exception:
        object.__setattr__(action, "validation", report)
    return action


def build_unknown_action(name: str, arguments: Any = None, *, action_id: Optional[str] = None) -> SingleAction:
    """Construct a best-effort unknown action marked invalid for name lookup paths."""
    parsed_args = _parse_arguments_payload(arguments)

    report = ValidationReport(
        valid=False,
        name_valid=False,
        args_valid=True,
        error="Unknown action",
        details={"reason": "unknown_action"},
    )
    payload: dict[str, Any] = {
        "name": name,
        "arguments": parsed_args if parsed_args is not None else {},
        "validation": report,
    }
    if action_id is not None:
        payload["id"] = action_id
    # model_construct to avoid BaseModel validation on arbitrary arguments types
    return SingleAction.model_construct(**payload)


def format_validation_errors(error: ValidationError) -> str:
    """Format pydantic validation errors into a compact human-readable string."""
    result = ""
    for err in error.errors():
        loc = ".".join(str(x) for x in err.get("loc", ()) if x is not Ellipsis) or "value"
        msg = err.get("msg", "Invalid value")
        input_value = err.get("input", None)
        input_type = type(input_value).__name__

        if err.get("type") == "missing":
            result += f"The field '{loc}' is required but was not provided."
        else:
            result += f"Field '{loc}': {msg} " f"(received {input_value!r} of type {input_type})."
    return result


def extract_argument(arguments: Any, field_name: str, default: Any = None) -> Any:
    """Best-effort extraction for a field from BaseModel/dict-like payloads."""
    if isinstance(arguments, BaseModel):
        return arguments.model_dump().get(field_name, default)
    if isinstance(arguments, dict):
        return arguments.get(field_name, default)
    return default


class ActionsHandler:
    """Central handler/registry for available actions and their handlers."""

    def __init__(
        self,
        logger: Optional[Logger] = None,
        *,
        warn_on_validation_error: bool = True,
        warn_on_unknown_action: bool = True,
        handle_validation_error: Optional[Callable[[SingleAction, str], Optional[SingleObservation]]] = None,
        handle_unknown_action: Optional[Callable[[SingleAction], Optional[SingleObservation]]] = None,
    ):
        if warn_on_unknown_action and handle_unknown_action is not None:
            raise ValueError("Cannot both warn and custom-handle unknown actions; set warn_on_unknown_action=False")
        if warn_on_validation_error and handle_validation_error is not None:
            raise ValueError("Cannot both warn and custom-handle validation errors; set warn_on_validation_error=False")
        self._logger = logger or get_disabled_logger()
        self._actions: dict[str, ActionType] = {}
        self._handlers: dict[str, ActionHandler] = {}
        self._warn_on_validation_error = warn_on_validation_error
        self._warn_on_unknown_action = warn_on_unknown_action
        self._handle_validation_error = handle_validation_error
        self._handle_unknown_action = handle_unknown_action or self._default_unknown_action
        self._stats: Counter[str] = Counter()

    # Registration ----------------------------------------------------------------
    def add_action(
        self,
        name: str,
        description: str,
        action_cls: type[SingleAction],
        handler: ActionHandler,
        *,
        is_finish: bool = False,
        is_message: bool = False,
        is_hidden: bool = False,
    ) -> ActionType:
        """Register a new action by specifying its parts; the ActionType is constructed internally."""
        action = ActionType(
            name=name,
            description=description,
            cls=action_cls,
            is_finish=is_finish,
            is_message=is_message,
            is_hidden=is_hidden,
        )
        self.add_action_type(action, handler)
        return action

    def add_action_type(self, action: ActionType, handler: ActionHandler) -> None:
        """Register an already-constructed ActionType."""
        if not isinstance(action, ActionType):
            raise ValueError("action must be an ActionType")
        self._store_action(action, handler)

    def add_actions(self, actions: list[ActionType], handler: ActionHandler) -> None:
        for action in actions:
            self.add_action_type(action, handler)

    # Accessors -------------------------------------------------------------------
    @property
    def actions(self) -> list[ActionType]:
        all_actions = list(self._actions.values())
        return list(filter(lambda action: not action.is_hidden, all_actions))

    def normalize(
        self,
        action: Optional[Action],
    ) -> Optional[SingleObservation | MultiObservation]:
        """Alias for execute() to emphasize validation/normalization use-cases."""
        return self.execute(action)

    # Execution -------------------------------------------------------------------
    def execute(
        self,
        action: Optional[Action],
    ) -> Optional[SingleObservation | MultiObservation]:
        """Execute user-supplied action(s) through registered handlers and return a merged observation."""
        if action is None:
            return None

        observations: list[SingleObservation] = []

        for single_action in action.to_action_list():
            outcome = self._execute_single(single_action)
            if outcome is not None:
                observations.append(outcome)

        if not observations:
            return None
        if len(observations) == 1:
            return observations[0]
        return MultiObservation(observations=observations)

    def _execute_single(self, action: SingleAction) -> Optional[SingleObservation]:
        handler = self._handlers.get(action.name)

        if handler is None:
            self._logger.error(f"Unknown action requested: {action.name}")
            self._record_error("unknown_action")
            return self._normalize_handler_result(self._handle_unknown_action(action), action)

        validation_error = self._validate_arguments(action)
        if validation_error:
            message = f"Validation Error in {action.name}: {validation_error}"
            self._logger.error(message)
            self._record_error("validation_error")
            if self._handle_validation_error is not None:
                return self._normalize_handler_result(
                    self._handle_validation_error(action, validation_error),
                    action,
                )
            if not self._warn_on_validation_error:
                return None
            return self._normalize_handler_result(message, action)

        try:
            raw_result = handler(action)
        except Exception as exc:  # pragma: no cover - defensive
            self._logger.exception(f"Action handler failed for {action.name}: {exc}")
            self._record_error("handler_exception")
            return self._normalize_handler_result(
                f"Action '{action.name}' failed: {exc}",
                action,
            )

        observation = self._normalize_handler_result(raw_result, action)
        return observation

    def _validate_arguments(self, action: SingleAction) -> Optional[str]:
        arguments = action.arguments
        expected_type = self._expected_arguments_type(action)

        report: Optional[ValidationReport] = action.validation
        if report:
            error = _validation_report_error(report)
            if error:
                return error
            return None

        # If arguments already a BaseModel, validate round-trip
        if isinstance(arguments, BaseModel):
            try:
                arguments.__class__.model_validate(arguments.model_dump())
            except ValidationError as e:
                return format_validation_errors(e)
            return None

        # If we know the expected type and got a dict, try to validate/construct
        if expected_type and issubclass(expected_type, BaseModel) and isinstance(arguments, dict):
            try:
                expected_type.model_validate(arguments)
            except ValidationError as e:
                return format_validation_errors(e)
            return None

        # Unknown/invalid argument shape
        if not isinstance(arguments, (BaseModel, dict)):
            msg = f"Invalid arguments type: {type(arguments).__name__}"
            if self._logger:
                self._logger.error(msg)
            self._record_error("invalid_arguments_type")
            return msg
        return None

    @staticmethod
    def _expected_arguments_type(action: SingleAction) -> Optional[type]:
        try:
            field = action.__class__.model_fields.get("arguments")
            if field and isinstance(field.annotation, type):
                return field.annotation  # type: ignore[return-value]
        except Exception:
            return None
        return None

    def _record_error(self, key: str) -> None:
        self._stats[key] += 1

    def get_errors_stats(self) -> dict[str, int]:
        return dict(self._stats)

    def _default_unknown_action(self, action: SingleAction) -> SingleObservation:
        if action.name == "message":
            text = "Error: Sending a message is not allowed. Please use only one of the available actions."
        else:
            text = f"Error: Unknown action - {action.name}"
        return SingleObservation(invoking_actions=[action], result=text)

    @staticmethod
    def _normalize_handler_result(raw_result: Any, action: SingleAction) -> Optional[SingleObservation]:
        if raw_result is None:
            return None
        if isinstance(raw_result, SingleObservation):
            if not raw_result.invoking_actions:
                raw_result.invoking_actions = [action]
            return raw_result
        return SingleObservation(invoking_actions=[action], result=raw_result)

    def _store_action(self, action: ActionType, handler: ActionHandler) -> None:
        self._validate_action_type(action)
        self._actions[action.name] = action
        self._handlers[action.name] = handler

    @staticmethod
    def _validate_action_type(action: ActionType) -> None:
        args_type = action.arguments
        if not isinstance(args_type, type) or not issubclass(args_type, BaseModel):
            raise ValueError(
                "Action arguments must be a Pydantic BaseModel. "
                f"Action '{action.name}' has arguments type {args_type!r}. "
                "Ensure the action's 'arguments' annotation resolves to a BaseModel."
            )



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import importlib
import os
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch


def test_dotenv_loads_settings_and_env_vars() -> None:
    with TemporaryDirectory() as tmpdir:
        env_path = Path(tmpdir) / ".env"
        env_path.write_text(
            "EXGENTIC_LOG_LEVEL=DEBUG\nWATSONX_API_KEY=abc123\n"  # pragma: allowlist secret
        )

        with patch.dict(os.environ, {"EXGENTIC_DOTENV_PATH": str(env_path)}, clear=False):
            os.environ.pop("EXGENTIC_LOG_LEVEL", None)
            os.environ.pop("WATSONX_API_KEY", None)

            sys.modules.pop("exgentic.utils.settings", None)
            settings_module = importlib.import_module("exgentic.utils.settings")

            settings_module.get_settings.cache_clear()
            settings = settings_module.get_settings()

            assert settings.log_level == "DEBUG"
            assert os.environ.get("WATSONX_API_KEY") == "abc123"
            assert settings.dotenv_path == str(env_path)



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import contextvars
import os
import shutil
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Iterator

from ..utils.paths import sanitize_path_component
from ..utils.settings import get_settings

# ---------------------------------------------------------------------------
# Role enum
# ---------------------------------------------------------------------------


class Role(str, Enum):
    FRAMEWORK = "framework"
    AGENT = "agent"
    BENCHMARK = "benchmark"


# ---------------------------------------------------------------------------
# OTEL Context dataclass
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class OtelContext:
    """OpenTelemetry span context for distributed tracing."""

    trace_id: str
    span_id: str


# ---------------------------------------------------------------------------
# Env-var keys used for subprocess transport
# ---------------------------------------------------------------------------

_ENV_RUN_ID = "EXGENTIC_CTX_RUN_ID"
_ENV_OUTPUT_DIR = "EXGENTIC_CTX_OUTPUT_DIR"
_ENV_CACHE_DIR = "EXGENTIC_CTX_CACHE_DIR"
_ENV_SESSION_ID = "EXGENTIC_CTX_SESSION_ID"
_ENV_TASK_ID = "EXGENTIC_CTX_TASK_ID"
_ENV_ROLE = "EXGENTIC_CTX_ROLE"
ENV_OTEL_TRACE_ID = "EXGENTIC_CTX_OTEL_TRACE_ID"
ENV_OTEL_SPAN_ID = "EXGENTIC_CTX_OTEL_SPAN_ID"
OTEL_ENABLED_ENV = "EXGENTIC_OTEL_ENABLED"


# ---------------------------------------------------------------------------
# Core Context dataclass
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class Context:
    run_id: str
    output_dir: str
    cache_dir: str
    session_id: str | None = None
    task_id: str | None = None
    role: Role = Role.FRAMEWORK
    otel_context: OtelContext | None = None

    def with_session(self, session_id: str, task_id: str | None = None) -> Context:
        return Context(
            run_id=self.run_id,
            output_dir=self.output_dir,
            cache_dir=self.cache_dir,
            session_id=session_id,
            task_id=task_id,
            role=self.role,
            otel_context=self.otel_context,
        )

    def with_role(self, role: Role) -> Context:
        return Context(
            run_id=self.run_id,
            output_dir=self.output_dir,
            cache_dir=self.cache_dir,
            session_id=self.session_id,
            task_id=self.task_id,
            role=role,
            otel_context=self.otel_context,
        )

    def with_otel_context(self, otel_context: OtelContext | None) -> Context:
        """Create a new Context with updated OTEL context."""
        return Context(
            run_id=self.run_id,
            output_dir=self.output_dir,
            cache_dir=self.cache_dir,
            session_id=self.session_id,
            task_id=self.task_id,
            role=self.role,
            otel_context=otel_context,
        )

    def to_env(self) -> dict[str, str]:
        env: dict[str, str] = {
            _ENV_RUN_ID: self.run_id,
            _ENV_OUTPUT_DIR: self.output_dir,
            _ENV_CACHE_DIR: self.cache_dir,
            _ENV_ROLE: self.role.value,
        }
        if self.session_id is not None:
            env[_ENV_SESSION_ID] = self.session_id
        if self.task_id is not None:
            env[_ENV_TASK_ID] = self.task_id
        if self.otel_context is not None:
            env[ENV_OTEL_TRACE_ID] = self.otel_context.trace_id
            env[ENV_OTEL_SPAN_ID] = self.otel_context.span_id
        return env

    @classmethod
    def from_env(cls, env: dict[str, str] | None = None) -> Context:
        src = env if env is not None else os.environ
        run_id = src.get(_ENV_RUN_ID, "")
        if not run_id:
            raise RuntimeError(f"{_ENV_RUN_ID} not set in environment.")
        run_id = sanitize_path_component(run_id)
        output_dir = src.get(_ENV_OUTPUT_DIR) or get_settings().output_dir
        cache_dir = src.get(_ENV_CACHE_DIR) or get_settings().cache_dir
        session_id = src.get(_ENV_SESSION_ID) or None
        task_id = src.get(_ENV_TASK_ID) or None
        role_str = src.get(_ENV_ROLE)
        try:
            role = Role(role_str) if role_str else Role.FRAMEWORK
        except ValueError:
            role = Role.FRAMEWORK

        # Read OTEL context if present
        otel_context: OtelContext | None = None
        trace_id = src.get(ENV_OTEL_TRACE_ID)
        span_id = src.get(ENV_OTEL_SPAN_ID)
        if trace_id and span_id:
            otel_context = OtelContext(trace_id=trace_id, span_id=span_id)

        return cls(
            run_id=run_id,
            output_dir=output_dir,
            cache_dir=cache_dir,
            session_id=session_id,
            task_id=task_id,
            role=role,
            otel_context=otel_context,
        )


# ---------------------------------------------------------------------------
# Single ContextVar — the single source of truth
# ---------------------------------------------------------------------------

_CONTEXT: contextvars.ContextVar[Context | None] = contextvars.ContextVar(
    "exgentic_context",
    default=None,
)

# Fallback for threads that don't inherit ContextVar (uvicorn thread-pool
# workers, service runner threads). Set by init_context_from_env() and
# set_context_fallback().
_SUBPROCESS_CONTEXT: Context | None = None

_ENV_LOCK = threading.Lock()


# ---------------------------------------------------------------------------
# Accessors
# ---------------------------------------------------------------------------


def get_context() -> Context:
    """Return the current Context. Raises RuntimeError if none is set."""
    ctx = _CONTEXT.get()
    if ctx is None:
        ctx = _SUBPROCESS_CONTEXT
    if ctx is None:
        raise RuntimeError("No context set. Use run_scope() or init_context_from_env().")
    return ctx


def try_get_context() -> Context | None:
    """Return the current Context, or None if none is set."""
    ctx = _CONTEXT.get()
    return ctx if ctx is not None else _SUBPROCESS_CONTEXT


def context_env() -> dict[str, str]:
    """Return context env vars for subprocess propagation, or empty dict."""
    ctx = try_get_context()
    if ctx is None:
        return {}
    return ctx.to_env()


@contextmanager
def context_env_scope() -> Iterator[None]:
    """Temporarily apply context env vars to os.environ (thread-safe)."""
    env = context_env()
    if not env:
        yield
        return
    with _ENV_LOCK:
        prev = {k: os.environ.get(k) for k in env}
        os.environ.update(env)
        try:
            yield
        finally:
            for k, v in prev.items():
                if v is None:
                    os.environ.pop(k, None)
                else:
                    os.environ[k] = v


def set_context(ctx: Context) -> None:
    """Imperatively set the current context."""
    _CONTEXT.set(ctx)


def set_context_fallback(ctx: Context | None) -> None:
    """Set a process-wide fallback for threads that don't inherit ContextVar."""
    global _SUBPROCESS_CONTEXT
    _SUBPROCESS_CONTEXT = ctx


# ---------------------------------------------------------------------------
# Context managers
# ---------------------------------------------------------------------------


@contextmanager
def run_scope(
    ctx: Context | None = None,
    *,
    run_id: str | None = None,
    output_dir: str | None = None,
    cache_dir: str | None = None,
    overwrite_run: bool = False,
) -> Iterator[Context]:
    """Enter a run context.

    Either pass an explicit *ctx*, or pass keyword args and the Context will
    be resolved from those args / env vars / settings defaults.
    """
    if ctx is None:
        ctx = _resolve_context(run_id, output_dir, cache_dir, overwrite_run)
    token = _CONTEXT.set(ctx)
    try:
        yield ctx
    finally:
        _CONTEXT.reset(token)


@contextmanager
def session_scope(session_id: str, task_id: str | None = None) -> Iterator[Context]:
    """Derive a session-scoped context from the current run context."""
    parent = get_context()
    ctx = parent.with_session(session_id, task_id)
    token = _CONTEXT.set(ctx)
    try:
        yield ctx
    finally:
        _CONTEXT.reset(token)


@contextmanager
def agent_scope() -> Iterator[Context]:
    """Set role=AGENT for the duration of the block, restore on exit."""
    parent = get_context()
    ctx = parent.with_role(Role.AGENT)
    token = _CONTEXT.set(ctx)
    try:
        yield ctx
    finally:
        _CONTEXT.reset(token)


@contextmanager
def benchmark_scope() -> Iterator[Context]:
    """Set role=BENCHMARK for the duration of the block, restore on exit."""
    parent = get_context()
    ctx = parent.with_role(Role.BENCHMARK)
    token = _CONTEXT.set(ctx)
    try:
        yield ctx
    finally:
        _CONTEXT.reset(token)


def init_context_from_env() -> Context:
    """Bootstrap ContextVar from env vars (called once in subprocess / Docker)."""
    global _SUBPROCESS_CONTEXT
    ctx = Context.from_env()
    _CONTEXT.set(ctx)
    _SUBPROCESS_CONTEXT = ctx
    return ctx


# ---------------------------------------------------------------------------
# Internals
# ---------------------------------------------------------------------------


def _resolve_context(
    run_id: str | None,
    output_dir: str | None,
    cache_dir: str | None,
    overwrite_run: bool,
) -> Context:
    settings = get_settings()
    resolved_run_id = run_id or os.environ.get(_ENV_RUN_ID) or datetime.now().isoformat().replace(":", "--")
    resolved_run_id = sanitize_path_component(resolved_run_id)
    resolved_output_dir = output_dir or os.environ.get(_ENV_OUTPUT_DIR) or settings.output_dir
    resolved_output_dir = str(Path(resolved_output_dir).resolve())
    resolved_cache_dir = cache_dir or os.environ.get(_ENV_CACHE_DIR) or settings.cache_dir
    resolved_cache_dir = str(Path(resolved_cache_dir).resolve())
    if overwrite_run:
        run_root = Path(resolved_output_dir) / resolved_run_id
        if run_root.exists():
            shutil.rmtree(run_root)
    return Context(
        run_id=resolved_run_id,
        output_dir=resolved_output_dir,
        cache_dir=resolved_cache_dir,
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
import logging
import subprocess
import textwrap
from pathlib import Path
from typing import Any, ClassVar, Literal

from pydantic import BaseModel, ConfigDict, Field

from ...core import Benchmark
from ...core.actions import ActionsHandler
from ...core.evaluator import Evaluator
from ...core.session import Session
from ...core.types import (
    Action,
    ActionType,
    BenchmarkResults,
    FinishAction,
    Observation,
    SessionIndex,
    SingleAction,
    SingleObservation,
)
from ...utils.logging import hook_loggers_into_session
from ...utils.paths import get_run_id
from ...utils.settings import ExgenticSettings, get_settings
from . import swebench_evaluation, swebench_logs, swebench_metrics

# =============================================================================
# Configuration
# =============================================================================

_CONFIG: dict[str, Any] = None


def get_config() -> dict[str, Any]:
    import yaml

    global _CONFIG
    if _CONFIG is None:
        path = Path(__file__).parent / "config.yaml"
        _CONFIG = yaml.safe_load(path.read_text())
    return _CONFIG


# =============================================================================
# Actions
# =============================================================================


class BashArgs(BaseModel):
    command: str = Field(description="Bash command to execute")


class BashAction(SingleAction):
    name: Literal["bash"] = "bash"
    arguments: BashArgs


class SubmitPatchArgs(BaseModel):
    summary: str = Field(description="Brief textual summary of the fix (no diff/patch)")


class SubmitPatchAction(FinishAction):
    name: Literal["finish"] = "finish"
    arguments: SubmitPatchArgs


# =============================================================================
# Action Handlers
# =============================================================================


def run_bash(
    command: str,
    env,
    timeout: int | None = None,
    size_limit: int | None = None,
    timeout_template: str = "",
) -> dict[str, Any]:
    """Execute bash command in environment."""
    try:
        output = env.execute(command=command, timeout=timeout or env.config.timeout)
    except subprocess.TimeoutExpired as e:
        exception_output = e.output.decode("utf-8", errors="replace") if e.output else ""
        msg = textwrap.dedent(timeout_template).format_map({"command": command, "exception_output": exception_output})
        output = {"output": msg, "returncode": 124}

    # Truncate large outputs
    if size_limit and len(output["output"]) > size_limit:
        head, tail = size_limit // 2, size_limit - size_limit // 2
        omitted = f"\n\n- OMITTED {len(output['output']) - size_limit} chars -\n\n"
        output["output"] = output["output"][:head] + omitted + output["output"][-tail:]

    return output


def generate_patch(env, cwd: str, base_commit: str) -> str:
    """Generate patch from staged changes."""
    command = f"git add -A && git diff --staged {base_commit} | cat"
    output = env.execute(command=command, cwd=cwd)
    return output["output"]


# =============================================================================
# Session
# =============================================================================


class SWEBenchSession(Session):
    """Session for a single SWE-bench task."""

    def __init__(
        self,
        settings: ExgenticSettings,
        instance: dict[str, Any],
        subset: str,
        max_interactions: int | None = None,
        require_submit_for_patch_evaluation: bool = True,
        session_id: str | None = None,
    ) -> None:
        cfg = get_config()

        self._instance = instance
        self._subset = subset
        self._instance_id = instance["instance_id"]
        if session_id is not None:
            self._session_id = session_id

        self._registry = ActionsHandler(
            logger=self.logger,
            warn_on_validation_error=False,
            warn_on_unknown_action=True,
            handle_validation_error=None,
            handle_unknown_action=None,
        )

        # State
        self._step_count = 0
        self._done = False
        self._max_interactions = max_interactions
        self._require_submit_for_patch_evaluation = require_submit_for_patch_evaluation
        self._action_count = 0
        self._score = None
        self._final_patch: str | None = None

        # Environment (set in start())
        self.env = None
        from swebench.harness.constants import DOCKER_WORKDIR

        self.container_repo_dir = DOCKER_WORKDIR
        self.container_base_commit: str | None = None

        # Config
        self._timeout = cfg["session"]["timeout"]
        self._environment_pull_timeout = int(cfg["session"].get("environment_pull_timeout", 600))
        self._observation_size_limit = cfg["session"]["observation_size_limit"]
        self._timeout_template = cfg["session"]["timeout_template"]
        self._task_prompt = cfg["task_prompt"]
        self._eval_config = cfg["evaluation"]

        self.logger.info(
            f"INIT | Session initialized | dataset: {self._subset:<30} "
            f"| instance_id: {self._instance_id:<20} | repo: {self._instance['repo']}"
        )

        hook_loggers_into_session(
            self.logger,
            logger_names=cfg["logging"].get("hook_loggers", []),
            level=logging._nameToLevel.get(cfg["logging"]["log_level"], logging.INFO),
        )

        # Call parent to save session manifest (session.json)
        super().__init__()

    # -------------------------------------------------------------------------
    # Lifecycle
    # -------------------------------------------------------------------------

    def start(self) -> Observation | None:
        self.logger.info("START | Session start")
        self._setup_environment()
        return SingleObservation(invoking_actions=[], result=None)

    def done(self) -> bool:
        return self._done

    def close(self):
        self.logger.info("CLOSE | Session closing...")
        super().close()
        self.logger.debug("CLOSE | Cleaning agent environment")
        del self.env
        self.logger.info("CLOSE | Session closed")

    # -------------------------------------------------------------------------
    # Step Execution
    # -------------------------------------------------------------------------

    def _handle_bash(self, action: BashAction) -> str:
        self.logger.info(f"STEP | {self._step_count:<3} | ACTION | bash | command: {action.arguments.command}")
        result = run_bash(
            command=action.arguments.command,
            env=self.env,
            timeout=self._timeout,
            size_limit=self._observation_size_limit,
            timeout_template=self._timeout_template,
        )
        self.logger.info(
            f"STEP | {self._step_count:<3} | RESULT | bash | "
            f"returncode: {result['returncode']} | output_len: {len(result['output'])}"
        )
        return result

    def _handle_submit_patch(self, action: SubmitPatchAction) -> str:
        self.logger.info(f"STEP | {self._step_count:<3} | ACTION | submit_patch | summary: {action.arguments.summary}")
        self._final_patch = generate_patch(
            env=self.env,
            cwd=self.container_repo_dir,
            base_commit=self.container_base_commit,
        )
        self._done = True
        return None

    def step(self, action: Action) -> Observation | None:
        if self._done:
            return None

        if self._max_interactions is not None:
            incoming = len(action.to_action_list())
            if self._action_count + incoming > self._max_interactions:
                self.logger.warning(
                    "STEP | {self._step_count:<3} | Max interactions reached (%s/%s); terminating session",
                    self._action_count,
                    self._max_interactions,
                )
                return None

        self._step_count += 1
        observation = self._registry.execute(action)
        if observation is None:
            return None

        # Post-execute: always track action count (useful for metrics)
        self._action_count += len(action.to_action_list())

        return observation

    # -------------------------------------------------------------------------
    # Scoring
    # -------------------------------------------------------------------------

    def score(self) -> swebench_logs.SessionScore:
        """Compute and cache the session score."""
        if self._score is not None:
            return self._score

        self.logger.info("SCORE | Computing session score")

        harness_result = None
        if self._done or not self._require_submit_for_patch_evaluation:
            if self._final_patch is None and not self._done:
                self.logger.info("SCORE | submit_patch not called; generating patch from current workspace")
                self._final_patch = self._generate_current_patch()
            harness_result = self._run_harness()
        else:
            self.logger.info(
                "SCORE | Skipping harness - submit_patch not called (require_submit_for_patch_evaluation=true)"
            )

        # Flush logs before parsing
        for handler in self.logger.handlers:
            handler.flush()

        # Build score
        self._score = swebench_logs.build_score(
            paths=self.paths,
            num_actions=self._action_count,
            max_interactions=self._max_interactions,
            instance_id=self._instance_id,
            harness_data=harness_result,
        )

        # Finished when agent called submit_patch (graceful completion)
        self._score.is_finished = self._done

        results_path = self.paths.benchmark_results
        swebench_logs.write_results(results_path, self._score)
        self.logger.info(
            "SCORE | Final | success=%s | score=%s | is_finished=%s",
            self._score.success,
            self._score.score,
            self._score.is_finished,
        )

        return self._score

    def _run_harness(self) -> swebench_evaluation.HarnessResult | None:
        """Run harness evaluation, handling errors."""
        if self._final_patch is None:
            self.logger.info("SCORE | Harness evaluation skipped: no patch available for evaluation")
            return None
        try:
            return swebench_evaluation.run_harness(
                patch=self._final_patch,
                instance_id=self._instance_id,
                subset=self._subset,
                paths=self.paths,
                eval_config=self._eval_config,
                logger=self.logger,
            )
        except Exception as e:
            self.logger.exception(f"SCORE | Harness evaluation failed: {e}")
            return None

    def _generate_current_patch(self) -> str | None:
        """Generate patch from current working tree for non-submit evaluation mode."""
        if self.env is None or self.container_base_commit is None:
            self.logger.warning("SCORE | Cannot generate patch: environment/base commit unavailable")
            return None
        try:
            return generate_patch(
                env=self.env,
                cwd=self.container_repo_dir,
                base_commit=self.container_base_commit,
            )
        except Exception as e:
            self.logger.exception(f"SCORE | Patch generation failed: {e}")
            return None

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def actions(self) -> list[ActionType]:
        if not self._registry.actions:
            self._registry.add_action(
                name="bash",
                description="Run a bash command in the repo root and get the output",
                action_cls=BashAction,
                handler=self._handle_bash,
            )
            self._registry.add_action(
                name="finish",
                description=(
                    "Finish the task by submitting a brief"
                    " summary. The system automatically computes"
                    " the git patch from the repository changes."
                ),
                action_cls=SubmitPatchAction,
                handler=self._handle_submit_patch,
                is_finish=True,
            )
        return self._registry.actions

    @property
    def task(self) -> str:
        return self._task_prompt.format_map(
            {
                "container_repo_dir": self.container_repo_dir,
                "problem_statement": self._instance["problem_statement"],
            }
        )

    @property
    def context(self) -> dict[str, str]:
        return {}

    @property
    def task_id(self) -> str:
        return str(self._instance_id)

    # -------------------------------------------------------------------------
    # Environment Setup
    # -------------------------------------------------------------------------

    def _setup_environment(self) -> None:
        """Initialize the Docker environment for the task."""
        self.logger.info("ENV | Setting up environment")
        from ...utils.logging import capture_stdio_to_session

        with capture_stdio_to_session(self.logger):
            import minisweagent
            import yaml
            from minisweagent.run.extra.swebench import get_sb_environment

            config_path = Path(minisweagent.__file__).parent / "config" / "extra" / "swebench.yaml"
            config = yaml.safe_load(config_path.read_text()).copy()
            config["environment"]["cwd"] = self.container_repo_dir
            config["environment"]["pull_timeout"] = self._environment_pull_timeout

            self.env = get_sb_environment(config=config, instance=self._instance)

            result = self.env.execute(command="git rev-parse HEAD", cwd=self.container_repo_dir)
            self.container_base_commit = result["output"].strip()

        if self.container_base_commit != self._instance["base_commit"]:
            self.logger.error(
                f"ENV | Base commit mismatch: expected {self._instance['base_commit']} "
                f"| got {self.container_base_commit}"
            )


# =============================================================================
# Evaluator
# =============================================================================


class SWEBenchEvaluator(Evaluator):
    """Evaluation logic for SWE-bench — task discovery, session config, aggregation."""

    def __init__(
        self,
        subset: str,
        require_submit_for_patch_evaluation: bool = True,
        max_interactions: int | None = None,
    ) -> None:
        self._subset = subset
        self._require_submit_for_patch_evaluation = require_submit_for_patch_evaluation
        self._max_interactions = max_interactions
        self._dataset: Any = None
        self._instances_by_id: dict[str, dict[str, Any]] = {}

    def list_tasks(self) -> list[str]:
        self._ensure_dataset()
        return list(self._instances_by_id.keys())

    def get_session_kwargs(self, index: SessionIndex) -> dict[str, Any]:
        self._ensure_dataset()
        task_id_str = str(index.task_id)
        instance = self._instances_by_id.get(task_id_str)
        if instance is None:
            raise KeyError(f"Unknown SWE-bench task id: {index.task_id}")
        return {
            "settings": get_settings(),
            "instance": instance,
            "subset": self._subset,
            "max_interactions": self._max_interactions,
            "require_submit_for_patch_evaluation": self._require_submit_for_patch_evaluation,
            "session_id": index.session_id,
        }

    def aggregate_sessions(self, sessions: list[SessionIndex]) -> BenchmarkResults:
        scores: list[float] = []
        session_ids = [s.session_id for s in sessions]
        for paths in self.get_sessions_paths(sessions):
            if not paths.benchmark_results.exists():
                raise FileNotFoundError(
                    f"Missing results for planned session '{paths.session_id}' at {paths.benchmark_results}"
                )
            with open(paths.benchmark_results, encoding="utf-8") as f:
                payload = json.load(f)
            scores.append(float(payload.get("score", 0.0)))

        metrics = swebench_metrics.collect_metrics(get_run_id(), session_ids)

        return BenchmarkResults(
            benchmark_name="swebench",
            total_tasks=len(sessions),
            score=sum(scores) / len(scores) if scores else 0.0,
            metrics=metrics["funnel"],
        )

    def _ensure_dataset(self) -> None:
        if self._dataset is not None:
            return
        if self._subset is None:
            raise ValueError("subset must be configured for SWE-bench.")
        from datasets import load_dataset

        dataset = load_dataset(self._subset, split="test")
        instances = list(dataset)
        if not instances:
            raise RuntimeError(
                f"SWE-bench dataset '{self._subset}' returned 0 instances. Check dataset availability and HF auth."
            )
        self._dataset = instances
        self._instances_by_id = {str(inst["instance_id"]): inst for inst in instances}


# =============================================================================
# Benchmark
# =============================================================================


class SWEBenchBenchmark(Benchmark):
    """Benchmark configuration for SWE-bench evaluation."""

    display_name: ClassVar[str] = "SWE-bench"
    slug_name: ClassVar[str] = "swebench"
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def get_evaluator_class(cls):
        return SWEBenchEvaluator

    @classmethod
    def get_session_class(cls):
        return SWEBenchSession

    subset: str | None = None
    require_submit_for_patch_evaluation: bool = True
    docker_socket: bool = True  # SWE-bench sessions create sibling Docker containers

    def model_post_init(self, __context):
        cfg = get_config()
        benchmark_cfg = cfg["benchmark"]
        session_cfg = cfg["session"]
        if self.subset is None:
            self.subset = benchmark_cfg["subset"]
        if self.runner is None:
            self.runner = benchmark_cfg.get("runner")
        if "seed" in benchmark_cfg:
            self.seed = benchmark_cfg["seed"]
        if (
            "require_submit_for_patch_evaluation" in benchmark_cfg
            and "require_submit_for_patch_evaluation" not in self.model_fields_set
        ):
            self.require_submit_for_patch_evaluation = bool(benchmark_cfg["require_submit_for_patch_evaluation"])
        if self.max_interactions is None:
            self.max_interactions = session_cfg.get("max_interactions")

    def get_evaluator_kwargs(self) -> dict[str, Any]:
        return {
            "subset": self.subset,
            "require_submit_for_patch_evaluation": self.require_submit_for_patch_evaluation,
            "max_interactions": self.max_interactions,
        }



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import os

from .health import acheck_model_accessible
from .proxy import LitellmProxy
from .trace_cost import load_trace_cost
from .trace_logger import (
    DEFAULT_FILE,
    FILE_ENV,
    TraceLogger,
    trace_logger,
)

# When running inside the LiteLLM proxy subprocess, eagerly initialise the
# Exgentic cache so that ``litellm.cache`` is set before any request arrives.
# The parent process sets EXGENTIC_PROXY_CACHE_INIT=true when it launches the
# proxy with disk caching enabled.
if os.environ.get("EXGENTIC_PROXY_CACHE_INIT", "").lower() in ("true", "1"):
    from ...utils.settings import get_settings

    get_settings()

__all__ = [
    "LitellmProxy",
    "trace_logger",
    "TraceLogger",
    "FILE_ENV",
    "DEFAULT_FILE",
    "load_trace_cost",
    "acheck_model_accessible",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

## CLI usage:
# exgentic evaluate --benchmark tau2 --agent openai_solo --subset retail --task 4 \
#   --model gpt-4o --set benchmark.user_simulator_model=gpt-4o
## Python API usage:
# from exgentic import RunConfig, evaluate
# evaluate(RunConfig(benchmark="tau2", agent="openai_solo", subset="retail", task_ids=["4"],
#   model="gpt-4o",
#   benchmark_kwargs={"user_simulator_model": "gpt-4o"}))
## Direct class usage (this script):
# TAU2Benchmark + OpenAIMCPAgent

from exgentic import OpenAIMCPAgent, TAU2Benchmark, evaluate


def main() -> None:
    benchmark = TAU2Benchmark(
        subset="retail",
        user_simulator_model="gpt-4o",
    )
    agent = OpenAIMCPAgent(model="gpt-4o")
    evaluate(
        benchmark=benchmark,
        agent=agent,
        output_dir="./outputs",
        task_ids=["4"],
    )


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json

from transformers import AutoTokenizer


class BCPSearchToolHandler:
    def __init__(
        self,
        searcher,
        snippet_max_tokens: int | None = None,
        k: int = 5,
        include_get_document: bool = True,
        full_doc_max_tokens: int | None = None,
    ):
        self.searcher = searcher
        self.snippet_max_tokens = snippet_max_tokens
        self.k = k
        self.include_get_document = include_get_document

        self.tokenizer = None
        self.full_doc_max_tokens = None
        if snippet_max_tokens and snippet_max_tokens > 0:
            self.tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen3-0.6B")
            self.full_doc_max_tokens = full_doc_max_tokens

    def execute_tool(self, tool_name: str, arguments: dict):
        if tool_name == "search":
            return self._search(arguments["query"])
        if tool_name == "get_document":
            return self._get_document(arguments["docid"])
        raise ValueError(f"Unknown tool: {tool_name}")

    def _search(self, query: str):
        candidates = self.searcher.search(query, self.k)

        if self.snippet_max_tokens and self.snippet_max_tokens > 0 and self.tokenizer:
            for cand in candidates:
                text = cand["text"]
                cand["snippet"] = self._truncate_text(text, self.snippet_max_tokens)
        else:
            for cand in candidates:
                cand["snippet"] = cand["text"]

        results = []
        for cand in candidates:
            if cand.get("score") is None:
                results.append({"docid": cand["docid"], "snippet": cand["snippet"]})
            else:
                results.append(
                    {
                        "docid": cand["docid"],
                        "score": cand["score"],
                        "snippet": cand["snippet"],
                    }
                )

        return json.dumps(results, indent=2)

    def _get_document(self, docid: str):
        try:
            result = self.searcher.get_document(docid)
        except Exception:
            result = None
        if result is None:
            return json.dumps({"error": f"Document {docid} not found"})

        text = result.get("text")
        result["text"] = self._truncate_text(text, max_len=self.full_doc_max_tokens)
        return json.dumps(result, indent=2)

    def _truncate_text(self, text: str, max_len: int) -> str:
        if not self.tokenizer:
            raise RuntimeError("Tokenizer not initialized")
        tokens = self.tokenizer.encode(text, add_special_tokens=False)
        if max_len and len(tokens) > max_len:
            truncated_tokens = tokens[:max_len]
            return self.tokenizer.decode(truncated_tokens, skip_special_tokens=True)
        return text



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import exgentic

from tests.api.fixtures.test_agent import TestAgent
from tests.api.fixtures.test_benchmark import TestBenchmark


def test_top_level_registry_class_imports_via_from_import():
    scope: dict[str, object] = {}
    exec("from exgentic import TestAgent, TestBenchmark", {}, scope)
    assert scope["TestAgent"] is TestAgent
    assert scope["TestBenchmark"] is TestBenchmark


def test_top_level_registry_class_imports_via_attribute_access():
    assert exgentic.TestAgent is TestAgent
    assert exgentic.TestBenchmark is TestBenchmark


def test_unknown_top_level_export_raises_attribute_error():
    try:
        _ = exgentic.NotARealExport
    except AttributeError as exc:
        assert "NotARealExport" in str(exc)
    else:
        raise AssertionError("Expected AttributeError for unknown top-level export.")



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Process-specific tests that verify actual process isolation."""

from __future__ import annotations

import gc
import os

import pytest
from exgentic.adapters.runners import with_runner

from .conftest import Calculator


def _make_process_calc(**kwargs):
    try:
        return with_runner(Calculator, runner="process", **kwargs)
    except PermissionError:
        pytest.skip("multiprocessing semaphores not available")


def test_runs_in_different_pid():
    calc = _make_process_calc(value=0)
    try:
        assert calc.pid() != os.getpid()
    finally:
        calc.close()


def test_crash_isolation():
    """Errors in the remote process don't crash the proxy."""
    calc = _make_process_calc(value=0)
    try:
        with pytest.raises(ZeroDivisionError):
            calc.divide(1, 0)
        assert calc.add(1, 2) == 3
    finally:
        calc.close()
        # Force cleanup of multiprocessing resources to avoid interference
        # with subsequent thread tests (CPython 3.11 bug workaround).
        gc.collect()



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import logging

import rich_click as click

from ..options import apply_debug_mode

logger = logging.getLogger(__name__)


@click.command("serve")
@click.option("--host", default="0.0.0.0", help="Host to bind to")
@click.option("--port", type=int, default=8080, help="Port to listen on")
@click.option("--cls", required=True, help="Class to serve (module.path:ClassName)")
@click.option("--kwargs", "kwargs_json", default=None, help="JSON constructor kwargs")
@click.option("--kwargs-b64", "kwargs_b64", default=None, help="Base64-encoded cloudpickle kwargs")
@click.option("--debug", is_flag=True, hidden=True)
def serve_cmd(host: str, port: int, cls: str, kwargs_json: str | None, kwargs_b64: str | None, debug: bool) -> None:
    """Serve a class instance over HTTP."""
    apply_debug_mode(debug)

    import importlib
    import json
    import os

    from ....core.context import init_context_from_env

    try:
        init_context_from_env()
    except RuntimeError as exc:
        logger.warning("Context init failed: %s", exc)
        ctx_vars = {k: v for k, v in os.environ.items() if k.startswith("EXGENTIC_CTX")}
        logger.debug("Context env vars: %s", ctx_vars)

    from ....adapters.runners.service import serve

    # Import the target module FIRST so that package __init__.py files
    # (which may set environment variables like TAU2_DATA_DIR) run before
    # cloudpickle deserialization triggers transitive library imports.
    module_path, attr_name = cls.rsplit(":", 1)
    mod = importlib.import_module(module_path)
    klass = getattr(mod, attr_name)

    # Deserialize kwargs: JSON (preferred) or cloudpickle fallback.
    if kwargs_b64 is not None:
        import base64

        import cloudpickle as cp

        kw = cp.loads(base64.b64decode(kwargs_b64))
    elif kwargs_json is not None:
        kw = json.loads(kwargs_json)
    else:
        kw = {}

    obj = klass(**kw)
    serve(obj, host=host, port=port)



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import logging
import time
from datetime import timedelta

import httpx
import pytest
from exgentic.adapters.agents import mcp_server as mcp_srv
from exgentic.utils.sync import run_sync
from mcp.client.session import ClientSession
from mcp.client.streamable_http import streamable_http_client
from mcp.shared.exceptions import McpError


class FakeMCP:
    def __init__(self, *args, **kwargs) -> None:
        fake_ts = type(
            "FakeTransportSecurity",
            (),
            {"allowed_hosts": [], "allowed_origins": []},
        )()
        self.settings = type(
            "FakeSettings",
            (),
            {"streamable_http_path": "/mcp", "transport_security": fake_ts},
        )()
        return

    def add_tool(self, fn, **kwargs) -> None:
        return None

    def streamable_http_app(self):
        return object()


class FakeConfig:
    def __init__(self, app, host, port, log_config=None) -> None:
        self.app = app
        self.host = host
        self.port = port
        self.log_config = log_config


class FakeServer:
    def __init__(self, config) -> None:
        self.config = config
        self.should_exit = False

    def run(self, *args, **kwargs) -> None:
        while not self.should_exit:
            time.sleep(0.01)


def _patch_uvicorn(monkeypatch):
    monkeypatch.setattr(mcp_srv.uvicorn, "Config", FakeConfig)
    monkeypatch.setattr(mcp_srv.uvicorn, "Server", lambda cfg: FakeServer(cfg))
    monkeypatch.setattr(mcp_srv, "FastMCP", FakeMCP)
    monkeypatch.setattr(mcp_srv, "wait_for_tcp", lambda *args, **kwargs: None)

    def _noop_run_sync(coro, timeout=None):
        coro.close()
        return

    monkeypatch.setattr(mcp_srv, "run_sync", _noop_run_sync)
    monkeypatch.setattr(mcp_srv.socket, "socket", _fake_socket_factory())


def _fake_socket_factory():
    class FakeSocket:
        _next_port = 10000

        def __init__(self, *args, **kwargs) -> None:
            self._port = None

        def setsockopt(self, *args, **kwargs) -> None:
            return None

        def bind(self, addr) -> None:
            _, port = addr
            if port == 0:
                FakeSocket._next_port += 1
                port = FakeSocket._next_port
            self._port = port

        def listen(self, backlog) -> None:
            return None

        def getsockname(self):
            return ("127.0.0.1", self._port)

        def close(self) -> None:
            return None

    return FakeSocket


def test_mcp_server_start_stop(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server")
    server = mcp_srv.MCPServer(
        FakeMCP(),
        host="127.0.0.1",
        port=12345,
        log_dir=tmp_path,
        logger=logger,
    )

    server.start(timeout=1.0)
    assert server.server is not None
    assert server.thread is not None and server.thread.is_alive()

    server.stop(timeout=1.0)
    assert server.server.should_exit is True
    assert server.thread is None or not server.thread.is_alive()

    server.stop(timeout=1.0)


def test_mcp_server_start_stop_idempotent(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server.idempotent")
    server = mcp_srv.MCPServer(
        FakeMCP(),
        log_dir=tmp_path,
        logger=logger,
    )

    server.start(timeout=1.0)
    first_thread = server.thread
    server.start(timeout=1.0)
    assert server.thread is first_thread

    server.stop(timeout=1.0)
    server.stop(timeout=1.0)


def test_mcp_server_start_tcp_failure_stops(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server.tcp_fail")
    server = mcp_srv.MCPServer(
        FakeMCP(),
        log_dir=tmp_path,
        logger=logger,
    )

    def _raise_tcp(*args, **kwargs):
        raise TimeoutError("tcp timeout")

    monkeypatch.setattr(mcp_srv, "wait_for_tcp", _raise_tcp)

    with pytest.raises(TimeoutError, match="tcp timeout"):
        server.start(timeout=1.0)

    assert server.thread is None or not server.thread.is_alive()


def test_mcp_server_start_ping_failure_stops(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server.ping_fail")
    server = mcp_srv.MCPServer(
        FakeMCP(),
        log_dir=tmp_path,
        logger=logger,
    )

    def _raise_timeout(coro, timeout=None):
        coro.close()
        raise TimeoutError("ping timeout")

    monkeypatch.setattr(mcp_srv, "run_sync", _raise_timeout)

    with pytest.raises(TimeoutError, match="ping timeout"):
        server.start(timeout=1.0)

    assert server.thread is None or not server.thread.is_alive()


def test_mcp_server_allocates_unique_ports(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server.ports")

    ports = set()
    servers = []
    for _ in range(50):
        server = mcp_srv.MCPServer(
            FakeMCP(),
            log_dir=tmp_path,
            logger=logger,
        )
        servers.append(server)
        server.start(timeout=1.0)
        ports.add(server.port)

    assert len(ports) == len(servers)
    for server in servers:
        server.stop(timeout=1.0)


def test_mcp_server_start_timeout(tmp_path, monkeypatch):
    _patch_uvicorn(monkeypatch)
    logger = logging.getLogger("test.mcp_server.timeout")
    server = mcp_srv.MCPServer(
        FakeMCP(),
        host="127.0.0.1",
        port=12345,
        log_dir=tmp_path,
        logger=logger,
    )

    def _no_start():
        time.sleep(0.2)

    monkeypatch.setattr(server, "_thread_entry", _no_start)

    with pytest.raises(RuntimeError, match="did not signal startup"):
        server.start(timeout=0.05)

    if server.thread and server.thread.is_alive():
        server.thread.join(timeout=0.5)


async def _call_tool(app, name: str, arguments: dict, read_timeout_seconds=None):
    async with app.router.lifespan_context(app):
        transport = httpx.ASGITransport(app=app)
        base_url = "http://127.0.0.1:9999"
        async with httpx.AsyncClient(
            transport=transport,
            base_url=base_url,
        ) as client:
            async with streamable_http_client(
                f"{base_url}/mcp",
                http_client=client,
            ) as (read_stream, write_stream, _):
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    return await session.call_tool(
                        name,
                        arguments=arguments,
                        read_timeout_seconds=read_timeout_seconds,
                    )


def test_mcp_server_unknown_tool_returns_mcp_error(tmp_path):
    logger = logging.getLogger("test.mcp_server.unknown_tool")

    def echo(text: str) -> str:
        return text

    server = mcp_srv.MCPServer(
        tools=[echo],
        log_dir=tmp_path,
        logger=logger,
    )
    app = server.mcp.streamable_http_app()

    result = run_sync(_call_tool(app, "does_not_exist", {}))
    assert result.isError is True
    assert any(block.text == "Unknown tool: does_not_exist" for block in result.content)


def test_mcp_server_invalid_tool_argument_name(tmp_path):
    logger = logging.getLogger("test.mcp_server.bad_arg_name")

    def echo(text: str) -> str:
        return text

    server = mcp_srv.MCPServer(
        tools=[echo],
        log_dir=tmp_path,
        logger=logger,
    )
    app = server.mcp.streamable_http_app()

    result = run_sync(_call_tool(app, "echo", {"wrong": "hi"}))
    assert result.isError is True
    assert any(
        "Error executing tool echo:" in block.text and "validation error" in block.text.lower()
        for block in result.content
    )


def test_mcp_server_invalid_tool_argument_type(tmp_path):
    logger = logging.getLogger("test.mcp_server.bad_arg_type")

    def echo(text: str) -> str:
        return text

    server = mcp_srv.MCPServer(
        tools=[echo],
        log_dir=tmp_path,
        logger=logger,
    )
    app = server.mcp.streamable_http_app()

    result = run_sync(_call_tool(app, "echo", {"text": 123}))
    assert result.isError is True
    assert any(
        "Error executing tool echo:" in block.text and "validation error" in block.text.lower()
        for block in result.content
    )


def test_mcp_server_call_tool_timeout(tmp_path):
    logger = logging.getLogger("test.mcp_server.call_timeout")

    def slow(text: str) -> str:
        time.sleep(0.2)
        return text

    server = mcp_srv.MCPServer(
        tools=[slow],
        log_dir=tmp_path,
        logger=logger,
    )
    app = server.mcp.streamable_http_app()

    timeout = timedelta(milliseconds=10)
    exc: BaseException | None = None
    try:
        run_sync(_call_tool(app, "slow", {"text": "hi"}, timeout))
    except BaseExceptionGroup as err:
        exc = err
    except McpError as err:
        exc = err

    assert exc is not None

    def _find_mcp_error(err: BaseException) -> McpError | None:
        if isinstance(err, McpError):
            return err
        if isinstance(err, BaseExceptionGroup):
            for sub in err.exceptions:
                found = _find_mcp_error(sub)
                if found is not None:
                    return found
        return None

    mcp_error = _find_mcp_error(exc)
    assert mcp_error is not None
    assert "Timed out while waiting for response" in str(mcp_error)



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

# Re-export from the installed package so existing tests keep working.
from exgentic.testing.benchmark import (
    TestBenchmark,
    TestEvaluator,
    TestSession,
)

__all__ = [
    "TestBenchmark",
    "TestEvaluator",
    "TestSession",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from typing import Any

from nicegui import ui

_STATUS_ORDER = [
    "success",
    "unsuccessful",
    "unfinished",
    "agent error",
    "benchmark error",
    "cancelled",
    "error",
    "running",
]

_STATUS_COLORS = {
    "success": "#22c55e",
    "unsuccessful": "#f59e0b",
    "unfinished": "#facc15",
    "agent error": "#ef4444",
    "benchmark error": "#f97316",
    "cancelled": "#94a3b8",
    "error": "#dc2626",
    "running": "#38bdf8",
}


def _status_from_outcome(success: Any, is_finished: Any, error_source: Any = None) -> str:
    if success is True:
        return "success"
    if is_finished is True:
        return "unsuccessful"
    if is_finished is False:
        return "unfinished"
    if error_source == "agent":
        return "agent error"
    if error_source == "benchmark":
        return "benchmark error"
    if error_source == "cancelled":
        return "cancelled"
    return "error"


def _status_counts_from_sessions(sessions: dict) -> dict[str, int]:
    counts: dict[str, int] = {}
    for data in sessions.values():
        status = data.get("status") or "error"
        counts[status] = counts.get(status, 0) + 1
    return counts


def _render_status_pie(status_counts: dict[str, int]) -> None:
    if not status_counts:
        ui.label("No session data available.")
        return
    data = []
    colors = []
    for status in _STATUS_ORDER:
        count = status_counts.get(status, 0)
        if count:
            data.append({"value": count, "name": status})
            colors.append(_STATUS_COLORS.get(status, "#94a3b8"))
    if not data:
        ui.label("No session data available.")
        return
    ui.echart(
        {
            "tooltip": {"trigger": "item"},
            "legend": {"orient": "vertical", "left": "left"},
            "color": colors,
            "series": [
                {
                    "name": "Sessions",
                    "type": "pie",
                    "radius": ["35%", "70%"],
                    "center": ["60%", "55%"],
                    "label": {"formatter": "{b}: {c} ({d}%)"},
                    "data": data,
                }
            ],
        }
    ).classes("w-full").style("height: 260px;")



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import os
from typing import Any, ClassVar

from ....core.types import ModelSettings
from ..base import ExecutionBackend, ProxyBackedAgent, ProxyBackedMCPAgentInstance
from .cli import ClaudeCLIConfig, ClaudeCodeCLI


class ClaudeCodeAgentInstance(ProxyBackedMCPAgentInstance):
    """Self-contained Claude Code CLI agent routed through a LiteLLM proxy."""

    def __init__(
        self,
        session_id: str,
        model_id: str,
        max_steps: int = 150,
        execution_backend: ExecutionBackend = ExecutionBackend.AUTO,
        model_settings: ModelSettings | None = None,
    ):
        # The alias is what we ask Claude Code CLI for; the proxy maps it to the backend model.
        self._claude_model_alias = "claude-3-5-sonnet-20241022"
        super().__init__(
            session_id,
            model_id,
            max_steps=max_steps,
            model_alias=self._claude_model_alias,
            execution_backend=execution_backend,
            model_settings=model_settings,
        )
        self._claude_log = self.paths.agent_dir / "claude_cli.log"

    @property
    def cli_display_name(self) -> str:
        return "Claude Code CLI"

    def _build_cli(self) -> ClaudeCodeCLI:
        cfg_dir = self.paths.agent_dir / "claude_code_config"
        return ClaudeCodeCLI(
            env=os.environ.copy(),
            log_path=self._claude_log,
            config_dir=cfg_dir,
            logger=self.logger,
            runner=self.execution_backend,
        )

    def _run_cli(
        self,
        cli: ClaudeCodeCLI,
        prompt: str,
        mcp_host: str,
        mcp_port: int,
        proxy: Any,
    ) -> Any:
        # allowed_tools = [f"mcp__environment__{action.name}" for action in self.actions]
        config = ClaudeCLIConfig(
            mcp_host=mcp_host,
            mcp_port=mcp_port,
            provider_url=proxy.base_url,
            backend_model=self.model_id,
            claude_model=self._claude_model_alias,
            # allowed_tools=allowed_tools,
            max_turns=self.max_steps,
        )
        config.env = {
            "MCP_TIMEOUT": str(config.mcp_timeout_ms),
            "MCP_TOOL_TIMEOUT": str(config.mcp_tool_timeout_ms),
        }
        result = cli.run(prompt=prompt, config=config)
        return result.stdout

    def _stringify_empty_output(self) -> bool:
        return True


class ClaudeCodeAgent(ProxyBackedAgent):
    display_name: ClassVar[str] = "Claude Code CLI"
    slug_name: ClassVar[str] = "claude_code"
    execution_backend: ExecutionBackend = ExecutionBackend.AUTO

    @classmethod
    def get_instance_class(cls):
        return ClaudeCodeAgentInstance

    def get_models_names(self) -> list[str]:  # type: ignore[override]
        return [str(self.model_id)]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""End-to-end evaluate flow through every runner.

Exercises the full evaluate() pipeline — evaluator creation, session execution,
scoring, close(), and aggregation — through each transport layer.

This catches issues like:
- ContextVar propagation across uvicorn thread-pool workers (service runner)
- Path resolution mismatches between host and runner
- close() forwarding through proxied objects
- Result file persistence across runner boundaries
- Volume mount and env var consistency (docker runner)
"""

from __future__ import annotations

import json
import platform
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import pytest
from exgentic import evaluate, execute
from exgentic.core.types import RunConfig

_RUNNERS = ["direct", "thread", "process", "service"]

# Detect Docker availability for conditional tests.
_docker_available = shutil.which("docker") is not None
if _docker_available:
    try:
        subprocess.run(["docker", "info"], check=True, capture_output=True, timeout=5)
    except Exception:
        _docker_available = False

if _docker_available and sys.version_info[:2] == (3, 12):
    _RUNNERS.append("docker")


@pytest.fixture(params=_RUNNERS)
def runner(request):
    return request.param


@pytest.fixture(scope="module")
def _docker_tmpdir():
    """Module-scoped temp dir for Docker tests under $HOME.

    Rancher Desktop / Docker Desktop on macOS only share ``/Users/`` by
    default.  pytest's ``tmp_path`` lives under ``/var/folders/`` which
    is NOT shared, so Docker volume mounts silently fail.

    Using a module-scoped parent dir avoids cleaning up temp dirs between
    tests, which would break stale logging FileHandlers held by the
    evaluate framework.
    """
    if platform.system() != "Darwin":
        yield None
        return
    d = Path(tempfile.mkdtemp(prefix=".exgentic_test_", dir=Path.home()))
    yield d
    shutil.rmtree(d, ignore_errors=True)


@pytest.fixture
def run_tmp(runner, tmp_path, _docker_tmpdir):
    """Temp path that works with Docker volume mounts on macOS."""
    if runner != "docker" or _docker_tmpdir is None:
        return tmp_path
    d = Path(tempfile.mkdtemp(dir=_docker_tmpdir))
    return d


def _run_config(tmp_path, runner: str, *, num_tasks: int = 2, policy: str = "good_then_finish") -> RunConfig:
    tasks = [f"task-{i}" for i in range(1, num_tasks + 1)]
    return RunConfig(
        benchmark="test_benchmark",
        agent="test_agent",
        output_dir=str(tmp_path / "outputs"),
        cache_dir=str(tmp_path / "cache"),
        run_id=f"run-{runner}",
        num_tasks=num_tasks,
        max_steps=10,
        max_actions=10,
        benchmark_kwargs={"tasks": tasks, "runner": runner},
        agent_kwargs={"policy": policy, "finish_after": 2},
    )


def test_evaluate_full_flow(run_tmp, runner):
    """Full evaluate (sessions + aggregation) through each runner."""
    config = _run_config(run_tmp, runner)
    results = evaluate(config)

    assert results.total_sessions == 2
    assert results.successful_sessions == 2
    assert results.benchmark_score == 1.0

    for session in results.session_results:
        assert session.score == 1.0
        assert session.success is True


def test_result_files_written(run_tmp, runner):
    """Verify all expected output files are created through each runner."""
    config = _run_config(run_tmp, runner, num_tasks=1)
    evaluate(config)

    run_dir = run_tmp / "outputs" / f"run-{runner}"
    session_id = config.to_session_config("task-1").get_session_id()
    session_dir = run_dir / "sessions" / session_id

    # Run-level files
    assert (run_dir / "results.json").exists()
    assert (run_dir / "benchmark_results.json").exists()

    # Session-level files
    assert (session_dir / "results.json").exists()
    assert (session_dir / "benchmark" / "results.json").exists()

    # Verify content is valid JSON with expected fields
    payload = json.loads((run_dir / "results.json").read_text())
    assert payload["total_sessions"] == 1
    assert payload["benchmark_score"] == 1.0


def test_execute_then_aggregate(run_tmp, runner):
    """Verify execute-only + aggregate-only works through each runner."""
    from exgentic import aggregate

    config = _run_config(run_tmp, runner, num_tasks=1)

    exec_results = execute(config)
    assert exec_results.benchmark_score is None
    assert exec_results.total_sessions == 1

    agg_results = aggregate(config)
    assert agg_results.benchmark_score == 1.0


def test_unsuccessful_session(run_tmp, runner):
    """Agent that finishes immediately scores 0 through each runner."""
    config = _run_config(run_tmp, runner, num_tasks=1, policy="finish_immediately")
    results = evaluate(config)

    assert results.total_sessions == 1
    assert results.successful_sessions == 0
    assert results.session_results[0].score == 0.0
    assert results.session_results[0].success is False


def test_parallel_workers(run_tmp, runner):
    """Evaluate with max_workers>1 to test thread-safety of pickling."""
    config = _run_config(run_tmp, runner, num_tasks=4)
    config = config.model_copy(update={"max_workers": 2})
    results = evaluate(config)

    assert results.total_sessions == 4
    assert results.successful_sessions == 4
    assert results.benchmark_score == 1.0



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import functools
import inspect
import logging
import socket
import threading
import time
from typing import Any, Callable, Iterable, Optional

import uvicorn
from mcp.client.session import ClientSession
from mcp.client.streamable_http import streamable_http_client
from mcp.server.fastmcp.server import FastMCP
from pydantic import BaseModel

from ...observers.logging import (
    configure_library_file_logging,
    configure_uvicorn_file_logging,
    get_logger,
)
from ...utils.sync import run_sync


class MCPServerConfig(BaseModel):
    http_timeout_seconds: float | None = None  # None means no timeout
    sse_read_timeout_seconds: float | None = None  # None means no timeout
    http_connect_timeout_seconds: float | None = None  # None means no timeout
    headers: dict[str, str] | None = None
    terminate_on_close: bool = True


class MCPServer:
    _MAX_SAFE_SCHEMA_INT = 2_147_483_647

    def __init__(
        self,
        mcp: FastMCP | None = None,
        *,
        host: str | None = None,
        port: int | None = None,
        tools: Iterable[Callable[..., Any]] | None = None,
        log_dir,
        logger: logging.Logger,
        stringify_empty_output: bool = False,
    ) -> None:
        self._mcp = mcp or self._build_fastmcp()
        self._host = host or "0.0.0.0"
        self._port = port
        self._log_dir = log_dir
        self._logger = logger
        self._stringify_empty_output = stringify_empty_output
        self._mcp_log_dir = self._log_dir / "mcp"
        self._mcp_log_dir.mkdir(parents=True, exist_ok=True)
        self._server_logger = get_logger(
            f"MCPServer_{id(self)}",
            str(self._mcp_log_dir / "server.log"),
        )
        ts = self._mcp.settings.transport_security
        ts.allowed_hosts = [
            *ts.allowed_hosts,
            "host.containers.internal:*",
            "host.docker.internal:*",
        ]
        ts.allowed_origins = [
            *ts.allowed_origins,
            "http://host.containers.internal:*",
            "http://host.docker.internal:*",
        ]

        tool_names: list[str] = []
        if tools:
            for fn in tools:
                tool_fn = self._wrap_tool(fn) if self._stringify_empty_output else fn
                tool = self._mcp._tool_manager.add_tool(tool_fn)
                tool.parameters = self._clamp_schema_ints(tool.parameters)
                tool_names.append(fn.__name__)
        self._log_tool_summary(tool_names)

        if self._logger is not None:
            self._logger.info("MCP logs at %s", self._mcp_log_dir)

        self._server: Optional[uvicorn.Server] = None
        self._thread: Optional[threading.Thread] = None
        self._sock: Optional[socket.socket] = None
        self._started = threading.Event()

    @classmethod
    def _clamp_schema_ints(cls, obj: Any) -> Any:
        if isinstance(obj, dict):
            return {k: cls._clamp_schema_ints(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [cls._clamp_schema_ints(v) for v in obj]
        if isinstance(obj, int) and not isinstance(obj, bool):
            if obj > cls._MAX_SAFE_SCHEMA_INT:
                return cls._MAX_SAFE_SCHEMA_INT
        if isinstance(obj, float):
            if obj > cls._MAX_SAFE_SCHEMA_INT:
                return float(cls._MAX_SAFE_SCHEMA_INT)
        return obj

    @staticmethod
    def stringify_empty_output(result: Any) -> Any:
        if result is None:
            return "null"
        if result == []:
            return "[]"
        return result

    @classmethod
    def _wrap_tool(cls, fn: Callable[..., Any]) -> Callable[..., Any]:
        signature = inspect.signature(fn)

        if inspect.iscoroutinefunction(fn):

            @functools.wraps(fn)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                result = await fn(*args, **kwargs)
                return cls.stringify_empty_output(result)

        else:

            @functools.wraps(fn)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                result = fn(*args, **kwargs)
                return cls.stringify_empty_output(result)

        wrapper.__signature__ = signature  # type: ignore[attr-defined]
        return wrapper

    def _build_fastmcp(self) -> FastMCP:
        root = logging.getLogger()
        null_handler: logging.Handler | None = None
        if not root.handlers:
            null_handler = logging.NullHandler()
            root.addHandler(null_handler)
        try:
            return FastMCP()
        finally:
            if null_handler is not None:
                try:
                    root.removeHandler(null_handler)
                except Exception:
                    pass

    @property
    def server(self) -> Optional[uvicorn.Server]:
        return self._server

    @property
    def thread(self) -> Optional[threading.Thread]:
        return self._thread

    @property
    def mcp(self) -> FastMCP:
        return self._mcp

    @property
    def host(self) -> str:
        return self._host

    @property
    def connect_host(self) -> str:
        # Use 127.0.0.1 for connecting, even if server binds to 0.0.0.0.
        return "127.0.0.1" if self._host == "0.0.0.0" else self._host

    @property
    def port(self) -> int:
        if self._port is None:
            raise RuntimeError("MCP server port not assigned yet.")
        return self._port

    def start(
        self,
        timeout: float = 60.0,
        *,
        tcp_timeout: float = 60.0,
        ping_timeout: float = 60.0,
    ) -> None:
        if self._thread and self._thread.is_alive():
            return

        self._ensure_socket()

        self._started.clear()
        t = threading.Thread(
            target=self._thread_entry,
            name=f"mcp-native:{self._port}",
            daemon=True,
        )
        self._thread = t
        t.start()

        if not self._started.wait(timeout=timeout):
            raise RuntimeError("MCP uvicorn thread did not signal startup within timeout.")

        try:
            started_at = time.time()
            # Use 127.0.0.1 for connecting, even if server binds to 0.0.0.0
            connect_host = "127.0.0.1" if self._host == "0.0.0.0" else self._host
            wait_for_tcp(connect_host, self.port, timeout=tcp_timeout)
            tcp_elapsed = time.time() - started_at
            ping_started = time.time()
            run_sync(
                wait_for_mcp_ping_async(connect_host, self.port, timeout=ping_timeout),
                timeout=ping_timeout + 5.0,
            )
            ping_elapsed = time.time() - ping_started
            self._server_logger.info(
                "MCP readiness OK (tcp=%.2fs, ping=%.2fs)",
                tcp_elapsed,
                ping_elapsed,
            )
        except BaseException as exc:
            self.stop(error=exc, raise_on_timeout=False)
            raise

    def __enter__(self) -> "MCPServer":
        self.start()
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.stop(error=exc, raise_on_timeout=True)

    def stop(
        self,
        timeout: float = 60.0,
        *,
        error: BaseException | None = None,
        raise_on_timeout: bool = True,
    ) -> None:
        if self._server is not None:
            self._server.should_exit = True

        t = self._thread
        if t and t.is_alive():
            t.join(timeout=timeout)
            if t.is_alive():
                message = "MCP uvicorn server thread did not exit cleanly"
                if raise_on_timeout and error is None:
                    raise RuntimeError(message)
                self._server_logger.warning(message)
        if self._sock is not None:
            try:
                self._sock.close()
            finally:
                self._sock = None

    def _thread_entry(self) -> None:
        thread_id = threading.get_ident()
        cleanup_uvicorn = configure_uvicorn_file_logging(
            self._mcp_log_dir / "uvicorn.log",
            thread_id=thread_id,
        )
        cleanup_mcp = configure_library_file_logging(
            self._mcp_log_dir / "server.log",
            logger_names=["mcp", "mcp.server", "mcp.client"],
            thread_id=thread_id,
        )

        server = self._build_server()

        self._server_logger.info(
            "Starting MCP server on %s:%s (path=%s)",
            self._host,
            self.port,
            self._mcp.settings.streamable_http_path,
        )
        self._started.set()

        try:
            sock = self._sock
            if sock is not None:
                self._server_logger.info("Using pre-bound socket on %s:%s", self._host, self.port)
                server.run(sockets=[sock])
            else:
                server.run()
        finally:
            self._server_logger.info("MCP server stopped")
            cleanup_mcp()
            cleanup_uvicorn()

    def _ensure_socket(self) -> None:
        if self._sock is not None:
            return
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((self._host, self._port or 0))
            sock.listen(1)
        except Exception:
            sock.close()
            raise
        self._sock = sock
        self._port = int(sock.getsockname()[1])

    def _build_server(self) -> uvicorn.Server:
        app = self._mcp.streamable_http_app()
        config = uvicorn.Config(app, host=self._host, port=self.port, log_config=None)
        server = uvicorn.Server(config)
        self._server = server
        return server

    def _log_tool_summary(self, tool_names: list[str]) -> None:
        count = len(tool_names)
        if count == 0:
            self._server_logger.info("Registered 0 MCP tools")
            return
        preview = ", ".join(tool_names[:6])
        if count > 6:
            preview = f"{preview}, +{count - 6} more"
        self._server_logger.info("Registered %s MCP tool(s): %s", count, preview)


def wait_for_tcp(host: str, port: int, timeout: float = 60.0) -> None:
    deadline = time.time() + timeout
    last_err: Optional[BaseException] = None
    while time.time() < deadline:
        try:
            with socket.create_connection((host, port), timeout=0.5):
                return
        except Exception as e:
            last_err = e
            time.sleep(0.1)
    raise TimeoutError(f"Server at {host}:{port} did not open TCP port within {timeout}s. " f"Last error: {last_err!r}")


async def wait_for_mcp_ping_async(host: str, port: int, timeout: float = 60.0) -> None:
    deadline = time.time() + timeout
    last_err: Optional[BaseException] = None
    url = f"http://{host}:{port}/mcp"

    import asyncio  # local import to keep module sync-first

    while time.time() < deadline:
        try:
            async with streamable_http_client(url) as (
                read_stream,
                write_stream,
                _,
            ):
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    await session.send_ping()
                    return
        except Exception as e:
            last_err = e
            await asyncio.sleep(0.1)

    raise TimeoutError(f"MCP ping at {host}:{port} failed within {timeout}s. " f"Last error: {last_err!r}")



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from pathlib import Path
from typing import Any, Optional

from nicegui import ui

from ....utils.paths import RunPaths
from ....utils.settings import get_settings
from .data import (
    _build_session_rows,
    _list_text_files,
    _load_text_file,
    load_run_config,
    load_session_file,
)
from .formatting import (
    _format_error_message,
    _format_payload,
    _format_value,
    _normalize_action_payload,
    _normalize_observation_payload,
)
from .state import SESSION_COLUMNS
from .status import _render_status_pie, _status_counts_from_sessions


def _metric_help_text(label: str) -> Optional[str]:
    return {
        "Benchmark": "Benchmark used for this run.",
        "Agent": "Agent used for this run.",
        "Models": "LLMs used in this run.",
        "Model": "Primary model used in this run.",
        "Subset": "Benchmark subset for this run.",
        "Running": "Currently running sessions / worker capacity.",
        "Completed": "Finished sessions / planned sessions.",
        "Total Sessions": "Total sessions recorded for this run.",
        "Successful": "Number of successful sessions.",
        "Benchmark Score": "Aggregate benchmark score for the run.",
        "Avg Score": "Average score across completed sessions.",
        "Avg Steps": "Average steps per session.",
        "Avg Agent Cost": "Average agent cost per session.",
        "Avg Benchmark Cost": "Average benchmark cost per session.",
        "Total Agent Cost": "Total agent cost across sessions.",
        "Total Benchmark Cost": "Total benchmark cost across sessions.",
        "Total Run Cost": "Total agent + benchmark cost.",
        "Finished": "Share of sessions marked finished.",
        "Success Rate": "Share of successful sessions.",
        "Finished Rate": "Share of sessions marked finished.",
        "Success %": "Share of successful sessions.",
        "Finished %": "Share of sessions marked finished.",
        "Status": (
            "Session status (running/success/unsuccessful/unfinished/" "agent error/benchmark error/cancelled/error)."
        ),
        "Steps": "Number of actions taken in the session.",
        "Score": "Session score from the benchmark.",
        "Exec Time (s)": "Elapsed time for the session.",
        "Agent Cost": "Cost attributed to the agent.",
        "Benchmark Cost": "Cost attributed to the benchmark.",
    }.get(label)


def _metric_info(help_text: str) -> None:
    ui.icon("info").classes("metric-info").tooltip(help_text)


def _metric_card(label: str, value: Any) -> None:
    with ui.element("div").classes("metric-card"):
        help_text = _metric_help_text(label)
        if help_text:
            _metric_info(help_text)
        ui.label(label).classes("metric-label")
        ui.label(_format_value(value)).classes("metric-value")


def _metric_card_small(label: str, value: Any) -> None:
    display_value = value if isinstance(value, str) else _format_value(value)
    with ui.element("div").classes("metric-card metric-card-sm"):
        help_text = _metric_help_text(label)
        if help_text:
            _metric_info(help_text)
        ui.label(label).classes("metric-label")
        ui.label(display_value).classes("metric-value")


def _metric_card_large(label: str, value: Any) -> None:
    display_value = value if isinstance(value, str) else _format_value(value)
    with ui.element("div").classes("metric-card metric-card-lg"):
        help_text = _metric_help_text(label)
        if help_text:
            _metric_info(help_text)
        ui.label(label).classes("metric-label")
        ui.label(display_value).classes("metric-value")


def _render_run_log(run_id: Optional[str], *, output_dir: Optional[str] = None) -> None:
    if not run_id:
        ui.label("No run log found.")
        return
    resolved_output = output_dir or get_settings().output_dir
    run_paths = RunPaths(run_id=run_id, output_dir=resolved_output)
    log_content = load_session_file(str(run_paths.tracker), "log")
    if log_content is None:
        legacy_path = run_paths.root / "tracker.log"
        if legacy_path != run_paths.tracker:
            log_content = load_session_file(str(legacy_path), "log")
    if log_content is None:
        ui.label("No run log found.")
        return
    ui.code(log_content, language="text").classes("w-full")


def _render_action_entry(payload: Any) -> None:
    data = _normalize_action_payload(payload)
    if "raw" in data:
        ui.code(_format_payload(data["raw"]), language="json").classes("w-full")
        return
    if "actions" in data:
        mode = str(data.get("mode", "parallel")).title()
        actions = data.get("actions") or []
        ui.label(f"{mode} actions ({len(actions)})").classes("metric")
        if not actions:
            ui.label("No actions provided.")
            return
        with ui.column().classes("w-full gap-3"):
            for idx, item in enumerate(actions, start=1):
                name = item.get("name") or f"Action {idx}"
                args = item.get("arguments")
                with ui.element("div").classes("trajectory-box"):
                    ui.label(name).classes("trajectory-title")
                    if args is None:
                        ui.label("No arguments.")
                    else:
                        ui.code(_format_payload(args), language="json").classes("w-full")
        return
    name = data.get("name") or "Action"
    args = data.get("arguments")
    if data.get("name") is None and data.get("arguments") is None:
        ui.label("Empty action.")
        return
    with ui.element("div").classes("trajectory-box"):
        ui.label(name).classes("trajectory-title")
        if args is None:
            ui.label("No arguments.")
        else:
            ui.code(_format_payload(args), language="json").classes("w-full")


def _render_observation_entry(payload: Any) -> None:
    data = _normalize_observation_payload(payload)
    if data is None:
        ui.label("Empty observation.")
        return
    if isinstance(data, list):
        ui.label(f"Observations ({len(data)})").classes("metric")
        if not data:
            ui.label("No observations provided.")
            return
        with ui.column().classes("w-full gap-3"):
            for item in data:
                if item is None:
                    ui.label("Empty observation.")
                    continue
                if isinstance(item, dict) and "sender" in item and "message" in item:
                    sender = item.get("sender") or "Sender"
                    message = item.get("message") or ""
                    with ui.element("div").classes("trajectory-box"):
                        ui.label(str(sender)).classes("trajectory-title")
                        ui.code(str(message), language="text").classes("w-full")
                    continue
                with ui.element("div").classes("trajectory-box"):
                    ui.code(_format_payload(item), language="json").classes("w-full")
        return
    if isinstance(data, dict) and "sender" in data and "message" in data:
        sender = data.get("sender") or "Sender"
        message = data.get("message") or ""
        with ui.element("div").classes("trajectory-box"):
            ui.label(str(sender)).classes("trajectory-title")
            ui.code(str(message), language="text").classes("w-full")
        return
    ui.code(_format_payload(data), language="json").classes("w-full")


def _render_trajectory_entry(kind: str, payload: Any) -> None:
    if kind == "action":
        _render_action_entry(payload)
        return
    if kind == "observation":
        _render_observation_entry(payload)
        return
    if kind == "error":
        ui.code(_format_payload(payload), language="json").classes("w-full")
        return
    ui.code(_format_payload(payload), language="json").classes("w-full")


def _render_overview_panel(
    sessions: dict,
    *,
    run_active: bool,
    run_id: Optional[str],
    results: Optional[dict] = None,
    planned_sessions: int | None = None,
    total_workers: int | None = None,
    run_meta: Optional[dict[str, Any]] = None,
    on_open_session=None,
) -> None:
    if run_meta:
        models = run_meta.get("models") or []
        models_text = ", ".join(models) if models else "-"
        with ui.element("div").classes("w-full metric-grid"):
            _metric_card_large("Benchmark", run_meta.get("benchmark"))
            _metric_card_large("Agent", run_meta.get("agent"))
            _metric_card_large("Models", models_text)
    if not sessions:
        if run_active:
            ui.label("Waiting for sessions...").classes("text-sm muted")
        else:
            ui.label("No sessions found.")
        return
    total = len(sessions)
    running = sum(1 for s in sessions.values() if s.get("status") == "running")
    done = total - running
    if planned_sessions is None:
        planned_sessions = total
    running_total = total_workers if total_workers is not None else "-"
    benchmark_score = results.get("benchmark_score") if isinstance(results, dict) else None
    with ui.element("div").classes("w-full metric-grid"):
        _metric_card("Running", f"{running}/{running_total}")
        _metric_card("Completed", f"{done}/{planned_sessions}")
        _metric_card("Benchmark Score", _format_value(benchmark_score))

    from .data import _build_overview_secondary_metrics

    secondary_metrics = _build_overview_secondary_metrics(sessions, results)
    with ui.element("div").classes("w-full metric-grid"):
        for label, value in secondary_metrics:
            _metric_card_small(label, value)

    with ui.card().classes("w-full card p-4"):
        ui.label("Status Breakdown").classes("section-title")
        status_counts = _status_counts_from_sessions(sessions)
        _render_status_pie(status_counts)

    with ui.card().classes("w-full card p-4"):
        rows = _build_session_rows(sessions)
        selection = "single" if on_open_session else None

        def _handle_select(e) -> None:
            selection_rows = getattr(e, "selection", None)
            if not selection_rows:
                return
            row = selection_rows[-1]
            if isinstance(row, dict):
                on_open_session(row)

        (
            ui.table(
                columns=SESSION_COLUMNS,
                rows=rows,
                row_key="session",
                selection=selection,
                on_select=_handle_select if on_open_session else None,
            )
            .classes("w-full")
            .props("flat")
        )


def _render_tabs(
    labels: list[str],
    renderers: dict[str, Any],
    *,
    default: Optional[str] = None,
) -> None:
    default_label = default or labels[0]
    with ui.tabs() as tabs:
        tab_by_name = {label: ui.tab(label) for label in labels}

    with ui.tab_panels(tabs, value=tab_by_name[default_label]).classes("w-full"):
        for label in labels:
            with ui.tab_panel(tab_by_name[label]):
                renderers[label]()


def _render_session_results_overview(meta: dict) -> None:
    with ui.element("div").classes("w-full metric-grid"):
        _metric_card("Status", meta.get("status", "-"))
        _metric_card("Steps", meta.get("steps", 0))
        _metric_card("Score", meta.get("score", "-"))
    error_msg = meta.get("error")
    if error_msg:
        ui.label("Error Details").classes("section-title")
        ui.code(_format_error_message(error_msg), language="text").classes("w-full").style("white-space: pre-wrap;")

    with ui.element("div").classes("w-full metric-grid"):
        _metric_card_small("Exec Time (s)", meta.get("execution_time"))
        _metric_card_small("Agent Cost", meta.get("agent_cost"))
        _metric_card_small("Benchmark Cost", meta.get("benchmark_cost"))


def _render_session_trajectory(turns_list: list[dict]) -> None:
    if not turns_list:
        ui.label("No trajectory data available yet.")
        return
    with ui.timeline(side="right"):
        for item in turns_list:
            kind = str(item.get("type", "event"))
            step_no = item.get("step", "?")
            content = item.get("content")
            title = kind.title()
            subtitle = f"Step {step_no}" if step_no != "?" else None
            icon = None
            if kind == "action":
                icon = "bolt"
            elif kind == "observation":
                icon = "visibility"
            elif kind == "error":
                icon = "error"
            with ui.timeline_entry(title=title, subtitle=subtitle, icon=icon):
                _render_trajectory_entry(kind, content)


def _render_session_logs(
    run_id: Optional[str],
    agent_files: list[Path],
    benchmark_files: list[Path],
) -> None:
    if not run_id:
        ui.label("No logs found.")
        return

    def _render_agent_logs() -> None:
        _render_log_files(agent_files)

    def _render_benchmark_logs() -> None:
        _render_log_files(benchmark_files)

    _render_tabs(
        ["Agent", "Benchmark"],
        {"Agent": _render_agent_logs, "Benchmark": _render_benchmark_logs},
    )


def _render_session_config(
    run_config_content: Optional[dict],
    benchmark_config_content: Optional[dict],
) -> None:
    if run_config_content is None and benchmark_config_content is None:
        ui.label("No config found.")
        return
    if run_config_content is not None:
        with ui.expansion("Run Config", value=False):
            ui.code(_format_payload(run_config_content), language="json").classes("w-full")
    if benchmark_config_content is not None:
        with ui.expansion("Benchmark Config", value=False):
            ui.code(_format_payload(benchmark_config_content), language="json").classes("w-full")


def _render_session_benchmark_results(results_content: Optional[dict]) -> None:
    if results_content is None:
        ui.label("No results found.")
    else:
        ui.code(_format_payload(results_content), language="json").classes("w-full")


def _render_session_tabs(
    *,
    meta: dict,
    session_data: Optional[dict],
    turns_list: list[dict],
    run_id: Optional[str],
    agent_files: list[Path],
    benchmark_files: list[Path],
    run_config_content: Optional[dict],
    benchmark_config_content: Optional[dict],
    results_content: Optional[dict],
) -> None:
    def _render_results_tab() -> None:
        _render_session_results_overview(meta)

    def _render_task_tab() -> None:
        _render_task_details(session_data)

    def _render_trajectory_tab() -> None:
        _render_session_trajectory(turns_list)

    def _render_logs_tab() -> None:
        _render_session_logs(run_id, agent_files, benchmark_files)

    def _render_config_tab() -> None:
        _render_session_config(run_config_content, benchmark_config_content)

    def _render_benchmark_results_tab() -> None:
        _render_session_benchmark_results(results_content)

    _render_tabs(
        ["Results", "Task", "Trajectory", "Logs", "Config", "Benchmark Results"],
        {
            "Results": _render_results_tab,
            "Task": _render_task_tab,
            "Trajectory": _render_trajectory_tab,
            "Logs": _render_logs_tab,
            "Config": _render_config_tab,
            "Benchmark Results": _render_benchmark_results_tab,
        },
        default="Results",
    )


def _render_dive_panel(
    sessions: dict,
    turns: dict,
    *,
    run_id: Optional[str],
    selected_session: Optional[str],
    on_session_change,
    output_dir: Optional[str] = None,
) -> None:
    session_ids = sorted(sessions.keys())
    if not session_ids:
        ui.label("No sessions to inspect yet.")
        return

    active_session = selected_session if selected_session in session_ids else session_ids[0]

    with ui.card().classes("w-full card p-4"):
        ui.select(session_ids, value=active_session, label="Select session").props("dense").on_value_change(
            on_session_change
        )
    meta = sessions.get(active_session, {})
    turns_list = turns.get(active_session, [])

    agent_files: list[Path] = []
    benchmark_files: list[Path] = []
    session_data = None
    run_config_content = None
    benchmark_config_content = None
    results_content = None
    if run_id:
        resolved_output = output_dir or get_settings().output_dir
        sess_paths = RunPaths(run_id=run_id, output_dir=resolved_output).session(active_session)
        base = sess_paths.benchmark_dir
        agent_files = _list_text_files(sess_paths.agent_dir)
        benchmark_files = _list_text_files(base)
        session_data = load_session_file(str(sess_paths.session_manifest), "json")
        run_config_content = load_run_config(run_id, output_dir=resolved_output)
        benchmark_config_content = load_session_file(str(base / "config.json"), "json")
        results_content = load_session_file(str(sess_paths.benchmark_results), "json")

    with ui.card().classes("w-full card p-4"):
        _render_session_tabs(
            meta=meta,
            session_data=session_data,
            turns_list=turns_list,
            run_id=run_id,
            agent_files=agent_files,
            benchmark_files=benchmark_files,
            run_config_content=run_config_content,
            benchmark_config_content=benchmark_config_content,
            results_content=results_content,
        )


def _render_log_files(files: list[Path]) -> None:
    if not files:
        ui.label("No logs found.")
        return
    for path in files:
        with ui.expansion(path.name, value=False):
            content = _load_text_file(path)
            if content is None:
                ui.label("Unable to read file.")
            else:
                language = "json" if path.suffix.lower() == ".json" else "text"
                ui.code(content, language=language).classes("w-full")


def _render_task_details(session_data: Optional[dict]) -> None:
    if not session_data:
        ui.label("No session data found.")
        return
    ordered_keys = ["task", "context", "actions"]
    seen = set()
    for key in ordered_keys:
        if key in session_data:
            seen.add(key)
            with ui.expansion(str(key), value=False):
                value = session_data.get(key)
                if isinstance(value, str):
                    ui.code(value, language="text").classes("w-full")
                else:
                    ui.code(_format_payload(value), language="json").classes("w-full")
    for key in sorted(session_data.keys()):
        if key in seen:
            continue
        with ui.expansion(str(key), value=False):
            value = session_data.get(key)
            if isinstance(value, str):
                ui.code(value, language="text").classes("w-full")
            else:
                ui.code(_format_payload(value), language="json").classes("w-full")



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import hashlib
from typing import Any

from ...utils.disk_cache import DiskCacheSessionMixin


class SearchDiskCacheSession(DiskCacheSessionMixin):
    CACHE_DIR = "./exgentic_session_cache/browsecomp/searcher"  # single DB for all runs

    def __init__(
        self,
        query: str,
        n: int,
        k: int,
        search_type: str,
        search_model: str,
        normalize: bool = False,
        use_cache: bool = True,
    ):
        self.query = query
        self.n = n
        self.k = k
        self.search_type = search_type
        self.search_model = search_model
        self.normalize = normalize
        self._init_cache_mixin(use_cache=use_cache)

    def build_cache_key_payload(self) -> dict:
        return {
            "search_type": self.search_type,
            "search_model": self.search_model,
            "normalize": self.normalize,
            "n": self.n,
            "k": self.k,
            "q": hashlib.sha256(self.query.encode("utf-8")).hexdigest(),
        }

    def build_additional_cache_metadata(self) -> dict:
        return {
            "search_type": self.search_type,
            "search_model": self.search_model,
            "normalize": self.normalize,
        }

    def on_cache_hit(self, payload: dict[str, Any]) -> bool:
        results = payload.get("results")
        if not isinstance(results, dict):
            return False
        result = results.get("raw")
        if result is None or not isinstance(result, str):
            return False
        return True

    def cache_results(self, result: str):
        self.set_results_payload({"raw": result})
        self.cache_score({})

    def handle_start_fetch_results(self):
        if self.handle_cache_start():
            cache_results = self.get_results_payload()
            if cache_results:
                try:
                    return cache_results["raw"]
                except Exception:
                    return None
        return None



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from exgentic.core.context import Context, set_context
from exgentic.integrations.litellm import LitellmProxy


def test_proxy_passes_context_env(monkeypatch):
    captured = {}

    class _DummyProc:
        def poll(self):
            return None

        def terminate(self) -> None:
            return None

        def wait(self, timeout=None) -> int:
            return 0

        def kill(self) -> None:
            return None

    def _fake_popen(*args, **kwargs):
        captured["env"] = kwargs.get("env", {})
        return _DummyProc()

    monkeypatch.setattr("exgentic.integrations.litellm.proxy.subprocess.Popen", _fake_popen)
    monkeypatch.setattr("exgentic.integrations.litellm.proxy._is_port_open", lambda *_a, **_k: True)
    monkeypatch.setattr("exgentic.integrations.litellm.proxy._is_proxy_ready", lambda *_a, **_k: True)

    ctx = Context(run_id="run-proxy", output_dir="/tmp/out", cache_dir="/tmp/cache")
    set_context(ctx)

    with LitellmProxy(model="openai/gpt-4o-mini", port=49998, startup_timeout=1.0):
        env = captured["env"]
        assert env["EXGENTIC_CTX_RUN_ID"] == "run-proxy"



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
import threading
import time
from collections import deque
from queue import Queue
from typing import Any

from pydantic import BaseModel

from ...core.agent import Agent
from ...core.context import get_context
from ...core.orchestrator.observer import Observer
from ...core.orchestrator.termination import (
    AgentError,
    BenchmarkError,
    RunCancelError,
    SessionCancelError,
)
from ...core.types import BenchmarkResults, SessionScore
from ...utils.paths import RunPaths


class DashboardEventsObserver(Observer):
    """Emit dashboard-friendly run events to a queue."""

    def __init__(self) -> None:
        self.events: Queue[dict[str, Any]] = Queue(maxsize=10000)
        self._thread_session: dict[int, str] = {}
        self._session_steps: dict[str, int] = {}
        self._session_started_at: dict[str, float] = {}
        self._session_agents: dict[str, Agent] = {}
        self._events_lock = threading.Lock()
        self._pending_events: deque = deque()
        self._last_batch_time = time.time()
        self._batch_interval = 0.02
        self._pending_run_meta = True

    def _emit(self, etype: str, **payload: Any) -> None:
        evt = {"type": etype, "ts": time.time(), **payload}

        current_time = time.time()

        if etype in ("session_started", "session_finished", "run_meta", "saved"):
            try:
                self.events.put_nowait(evt)
            except Exception:
                try:
                    self.events.get_nowait()
                    self.events.put_nowait(evt)
                except Exception:
                    pass
            return

        with self._events_lock:
            if len(self._pending_events) >= 100:
                self._pending_events.popleft()
            self._pending_events.append(evt)

            if current_time - self._last_batch_time >= self._batch_interval or len(self._pending_events) >= 10:
                self._flush_batch()
                self._last_batch_time = current_time

    def _flush_batch(self) -> None:
        while self._pending_events:
            try:
                evt = self._pending_events.popleft()
                self.events.put_nowait(evt)
            except Exception:
                try:
                    self.events.get_nowait()
                    self.events.put_nowait(evt)
                except Exception:
                    break

    def on_run_start(self, run_config) -> None:
        self._emit_run_meta()

    def on_session_start(self, session, agent: Agent, observation) -> None:
        sid = session.session_id
        tid = threading.get_ident()
        with self._events_lock:
            if tid is not None:
                self._thread_session[int(tid)] = sid
            self._session_steps[sid] = 0
            self._session_started_at[sid] = time.time()
            self._session_agents[sid] = agent
        self._emit("session_started", session_id=sid)

    def on_react_success(self, session, action) -> None:
        step_n = None
        sid = None
        tid = threading.get_ident()
        if tid is not None:
            with self._events_lock:
                sid = self._thread_session.get(int(tid))
                if sid:
                    step_n = self._session_steps.get(sid, 0) + 1
                    self._session_steps[sid] = step_n

        def _action_summary(a: Any) -> str:
            try:
                from ...core.types import ParallelAction, SingleAction

                if isinstance(a, SingleAction):
                    return f"{a.name}"[:100]
                if isinstance(a, ParallelAction):
                    return f"parallel[{len(a.actions)} actions]"
            except Exception:
                pass
            return str(a)[:100]

        def _action_obj(a: Any):
            try:
                from ...core.types import ParallelAction, SingleAction

                if isinstance(a, SingleAction):
                    args = a.arguments.model_dump()
                    if isinstance(args, dict) and len(str(args)) > 500:
                        args = {k: (v if len(str(v)) < 50 else f"{str(v)[:50]}...") for k, v in list(args.items())[:5]}
                    return {"type": "single", "name": a.name, "arguments": args}
                if isinstance(a, ParallelAction):
                    items = []
                    for i, x in enumerate(a.actions[:5]):
                        try:
                            args = x.arguments.model_dump()
                            if isinstance(args, dict) and len(str(args)) > 200:
                                args = {
                                    k: (v if len(str(v)) < 30 else f"{str(v)[:30]}...")
                                    for k, v in list(args.items())[:3]
                                }
                            items.append({"name": x.name, "arguments": args})
                        except Exception:
                            items.append(str(x)[:100])
                        if i >= 4:
                            break
                    return {"type": "parallel", "actions": items}
            except Exception:
                return None
            return None

        if sid:
            agent_cost, benchmark_cost = self._get_cost_snapshot(sid, session)
            execution_time = self._get_execution_time(sid)
            self._emit(
                "step",
                event="action",
                session_id=sid,
                n=step_n,
                action=_action_summary(action),
                action_obj=_action_obj(action),
                execution_time=execution_time,
                agent_cost=agent_cost,
                benchmark_cost=benchmark_cost,
            )

    def on_step_success(self, session, observation) -> None:
        sid = None
        tid = threading.get_ident()
        if tid is not None:
            with self._events_lock:
                sid = self._thread_session.get(int(tid))

        def _safe_json(obj: Any) -> str:
            try:
                if isinstance(obj, BaseModel):
                    data = obj.model_dump()
                    if isinstance(data, dict) and len(str(data)) > 1000:
                        truncated = {
                            k: (v if len(str(v)) < 100 else f"{str(v)[:100]}...") for k, v in list(data.items())[:10]
                        }
                        return json.dumps(truncated, ensure_ascii=False)
                    return json.dumps(data, ensure_ascii=False)
                obj_str = str(obj)
                if len(obj_str) > 1000:
                    obj_str = obj_str[:1000] + "..."
                return json.dumps(obj_str, default=str, ensure_ascii=False)
            except Exception:
                return str(obj)[:500]

        def _obs_obj(o: Any):
            try:
                if isinstance(o, BaseModel):
                    data = o.model_dump()
                    if isinstance(data, dict) and len(str(data)) > 1000:
                        truncated = {
                            k: (v if len(str(v)) < 100 else f"{str(v)[:100]}...") for k, v in list(data.items())[:10]
                        }
                        return truncated
                    return data
                safe_json = _safe_json(o)
                return json.loads(safe_json)
            except Exception:
                return None

        if sid:
            step_n = self._session_steps.get(sid, 0)
            agent_cost, benchmark_cost = self._get_cost_snapshot(sid, session)
            execution_time = self._get_execution_time(sid)
            self._emit(
                "observation",
                event="observation",
                session_id=sid,
                step=step_n,
                observation=_obs_obj(observation),
                initial=False,
                execution_time=execution_time,
                agent_cost=agent_cost,
                benchmark_cost=benchmark_cost,
            )

    def on_session_success(self, session, score: SessionScore, agent: Agent) -> None:
        sid = session.session_id
        self._flush_pending()
        self._cleanup_thread_session()
        success = bool(score.success)
        value = score.score
        details = score.model_dump()
        with self._events_lock:
            steps = self._session_steps.pop(sid, 0)
            started_at = self._session_started_at.pop(sid, None)
            self._session_agents.pop(sid, None)
        execution_time = time.time() - started_at if started_at is not None else None
        agent_cost = 0.0
        benchmark_cost = 0.0
        if agent is not None:
            try:
                report = agent.get_cost()
                agent_cost = float(report.total_cost)
            except Exception:
                agent_cost = 0.0
        try:
            report = session.get_cost()
            benchmark_cost = float(report.total_cost)
        except Exception:
            benchmark_cost = 0.0
        self._emit(
            "session_finished",
            session_id=sid,
            success=success,
            score=value,
            details=details,
            steps=steps,
            execution_time=execution_time,
            agent_cost=agent_cost,
            benchmark_cost=benchmark_cost,
            is_finished=score.is_finished,
        )

    def on_session_error(self, session, error) -> None:
        sid = session.session_id
        self._flush_pending()
        self._cleanup_thread_session()
        error_source = None
        if isinstance(error, AgentError):
            error_source = "agent"
        elif isinstance(error, BenchmarkError):
            error_source = "benchmark"
        elif isinstance(error, (SessionCancelError, RunCancelError, KeyboardInterrupt)):
            error_source = "cancelled"
        root_error = error.error if isinstance(error, (AgentError, BenchmarkError)) else None
        error_message = str(root_error) if root_error else str(error)
        details = {"error": error_message}
        if error_source is not None:
            details["error_source"] = error_source
        with self._events_lock:
            steps = self._session_steps.pop(sid, 0)
            started_at = self._session_started_at.pop(sid, None)
            self._session_agents.pop(sid, None)
        execution_time = time.time() - started_at if started_at is not None else None
        agent_cost = 0.0
        benchmark_cost = 0.0
        try:
            report = session.get_cost()
            benchmark_cost = float(report.total_cost)
        except Exception:
            benchmark_cost = 0.0
        self._emit(
            "session_finished",
            session_id=sid,
            success=False,
            score=None,
            details=details,
            steps=steps,
            execution_time=execution_time,
            agent_cost=agent_cost,
            benchmark_cost=benchmark_cost,
            is_finished=None,
            error_source=error_source,
        )

    def _get_execution_time(self, session_id: str) -> float | None:
        with self._events_lock:
            started_at = self._session_started_at.get(session_id)
        if started_at is None:
            return None
        return time.time() - started_at

    def _get_cost_snapshot(self, session_id: str, session) -> tuple[float, float]:
        with self._events_lock:
            agent = self._session_agents.get(session_id)
        agent_cost = 0.0
        benchmark_cost = 0.0
        if agent is not None:
            try:
                report = agent.get_cost()
                agent_cost = float(report.total_cost)
            except Exception:
                agent_cost = 0.0
        try:
            report = session.get_cost()
            benchmark_cost = float(report.total_cost)
        except Exception:
            benchmark_cost = 0.0
        return agent_cost, benchmark_cost

    def on_run_success(self, results: BenchmarkResults, run_config) -> None:
        payload = results.model_dump()
        self._emit("benchmark_recorded", results=payload)
        self._emit_saved_from_context()

    def on_run_error(self, error) -> None:
        self._emit_saved_from_context()

    def emit_saved(self, path: str) -> None:
        self._flush_pending()
        self._emit("saved", path=path)

    def _flush_pending(self) -> None:
        with self._events_lock:
            if self._pending_events:
                self._flush_batch()

    def _cleanup_thread_session(self) -> None:
        tid = threading.get_ident()
        if tid is not None:
            with self._events_lock:
                self._thread_session.pop(int(tid), None)

    def _emit_run_meta(self) -> None:
        if not self._pending_run_meta:
            return
        try:
            run_id = get_context().run_id
        except RuntimeError:
            return
        self._emit("run_meta", run_id=run_id)
        self._pending_run_meta = False

    def _emit_saved_from_context(self) -> None:
        try:
            ctx = get_context()
        except RuntimeError:
            return
        results_path = RunPaths.from_context(ctx).results
        self.emit_saved(str(results_path))



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Tau2 import shim that centralizes logging configuration.

This module suppresses Tau2's default console logging at import time and
provides re-exports for Tau2 symbols used by Exgentic. It also disables
propagation on the stdlib logger for Tau2 so library logs don't bubble to
the application's console. Session code can add file sinks as needed.
"""

from __future__ import annotations

import logging

# 1) Quiet stdlib logging for the 'tau2' namespace
_tau2_logger = logging.getLogger("tau2")
if not _tau2_logger.handlers:
    _tau2_logger.addHandler(logging.NullHandler())
_tau2_logger.propagate = False

# 2) Quiet Loguru's default console sink if Loguru is present
try:
    from loguru import logger as _loguru

    # Remove all default sinks to avoid console output. Session code will add
    # a file sink per session when needed.
    try:
        _loguru.remove()
    except Exception:
        pass
except Exception:
    _loguru = None  # type: ignore

# 3) Re-export Tau2 modules used by Exgentic
from rich.console import Console  # noqa: E402
from tau2.agent.llm_agent import LLMAgent  # noqa: E402
from tau2.data_model.message import (  # noqa: E402
    AssistantMessage,
    MultiToolMessage,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from tau2.data_model.simulation import Results, RunConfig, TerminationReason  # noqa: E402
from tau2.environment.tool import Tool  # noqa: E402
from tau2.metrics.agent_metrics import compute_metrics, is_successful  # noqa: E402
from tau2.registry import registry  # noqa: E402
from tau2.run import load_tasks, run_domain  # noqa: E402
from tau2.utils.display import ConsoleDisplay  # noqa: E402

# Tau2's llm_utils disables LiteLLM cache by default; re-enable Exgentic cache here.
from ...integrations.litellm.config import configure_litellm  # noqa: E402
from ...utils.settings import get_settings  # noqa: E402

configure_litellm(config=get_settings().to_litellm_config(), cache_only=True)

__all__ = [
    "AssistantMessage",
    "Console",
    "ConsoleDisplay",
    "LLMAgent",
    "MultiToolMessage",
    "Results",
    "RunConfig",
    "TerminationReason",
    "Tool",
    "ToolCall",
    "ToolMessage",
    "UserMessage",
    "compute_metrics",
    "is_successful",
    "load_tasks",
    "registry",
    "run_domain",
]
